# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from functools import partial
from inspect import signature
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
    Union,
    Type,
    Generator,
    Any
)
from lib389._constants import DIRSRV_STATE_ONLINE
from lib389.utils import ensure_bytes, ensure_str, ensure_list_str, ensure_int


DSLintSpec = Tuple[str, Callable]
DSLintParsedSpec = Tuple[Optional[str], Optional[str]]
DSLintClassSpec = Generator[DSLintSpec, None, None]
DSLintMethodSpec = Union[str, None, Type[List]]
DSLintResults = Generator[Any, None, None]


class DSLint():
    """In a super-class, create a method with name beginning with `_lint_`
    which would yield results (as described below). Such a method will
    then be available to the `lint()` method of the class.

    `lint_list`: takes a spec and yields available lints, recursively
    `lint`:      takes a spec and runs lints according to it, yielding results if any

    `spec`: is a colon-separated string, with prefix matching a method name and suffix
            being passed down to the method.

    A class inheriting from hereby class shall implement a method named `lint_uid()` which
    returns a pretty name of the object. This is to be used by a higher level code.

    Each lint method has to have a name prefix with _lint_. It may accept an optional
    parameter `spec` in which case:
    - it has to accept typing.List class as a parameter, in which case it shall yield
      all possible lint specs for that method
    - it receives the suffix provided to the `spec` of hereby `lint` method (as mentioned above)

    This means that we can detect and report common administrative errors
    in the server from our cli and rest tools.

    The structure of a result shall be:

        {
        dsle: '<identifier>'. dsle == ds lint error. Will be a code unique to
                            this module for the error, IE DSBLE0001.
        severity: '[HIGH:MEDIUM:LOW]'. severity of the error.
        items: '(dn,dn,dn)'. List of affected DNs or names.
        detail: 'msg ...'. An explination of the error.
        fix: 'msg ...'. Steps to resolve the error.
        }
    """

    @classmethod
    def _dslint_fname(cls, method: Callable) -> Optional[str]:
        """Return a pretty name for a method."""
        if callable(method) and method.__name__.startswith('_lint_'):
            return method.__name__[len('_lint_'):]
        else:
            return None

    @staticmethod
    def _dslint_parse_spec(spec: Optional[str]) -> DSLintParsedSpec:
        """Split `spec` to prefix and suffix."""
        wanted, *rest = spec.split(':', 1) if spec else (None, None)
        return (wanted if wanted not in [None, '*'] else None,
                rest[0] if rest else None)

    @classmethod
    def _dslint_make_spec(cls, method: Callable, spec: Optional[str] = None) -> str:
        """Build a new spec from prefix (`method` name) and suffix (`spec`)."""
        fname = cls._dslint_fname(method)
        return f'{fname}:{spec}' if spec else fname

    def lint_list(self, spec: Optional[str] = None) -> DSLintClassSpec:
        """Yield specs the object provides.

        This yields from each lint method yielding all specs it can provide.
        """

        assert hasattr(self, 'lint_uid')

        # Find _lint_ methods
        # NOTE: There is a caveat: don't you dare try to getattr on a @property, or
        #       you get it executed. That's why the following line's complexity.
        fs = [getattr(self, f) for f in dir(self)
              if f.startswith('_lint_') and self._dslint_fname(getattr(self, f))]

        # Filter acording to the `spec`
        wanted, rest = self._dslint_parse_spec(spec)
        if wanted:
            try:
                fs = [next(filter(lambda f: self._dslint_fname(f) == wanted, fs))]
            except StopIteration:
                raise ValueError('there is no such lint function')

        # Yield known specs
        for f in fs:
            fspec_t = signature(f).parameters.get('spec', None)
            if fspec_t:
                assert fspec_t.annotation == DSLintMethodSpec
                for fspec in [rest] if rest else f(spec=List):
                    yield self._dslint_make_spec(f, fspec), partial(f, spec=fspec)
            else:
                yield self._dslint_make_spec(f, rest), f

    def lint(self, spec: DSLintMethodSpec = None) -> DSLintResults:
        """Lint the object according to the `spec`."""

        if spec == List:
            yield from self.lint_list()
        else:
            for fn, f in self.lint_list(spec):
                yield from f()


class DSLints():
    """This is a meta class to provide lint functionality to classes that provide
    method `list` which returns list of objects that inherit from DSLint.

    Calling `lint` or `lint_list` method yields from respective object's methods.

    The `spec` is a colon-separated string. Its prefix matches the respective object's
    `lint_uid` (or all when asterisk); the suffix is passed down to the respective
    object's method.
    """

    def lint_list(self, spec: Optional[str] = None) -> DSLintClassSpec:
        """Yield specs the objects returned by `list` method provide."""

        assert hasattr(self, 'list')

        # Filter acording to the `spec`
        wanted, rest_spec = DSLint._dslint_parse_spec(spec)
        if wanted in [None, '*']:
            clss = self.list()
        else:
            clss = (cls for cls in self.list() if cls.lint_uid() == wanted)

        # Yield known specs
        for cls in clss:
            for fn, f in cls.lint_list(spec=rest_spec):
                yield (f'{cls.lint_uid()}:{fn}',
                       partial(f, rest_spec) if rest_spec else f)

    def lint(self, spec: DSLintMethodSpec = None) -> DSLintResults:
        """Lint the objects returned by `list` method according to the `spec`."""

        if spec == List:
            yield from self.lint_list()
        else:
            if ":" in spec:
                # Skip over any decorators, and get the actual function name
                check_name = spec.split(":")[-1]
            else:
                check_name = spec
            for obj in self.list():
                yield from obj.lint(check_name)


def _cached_dse_ldif_for_lint(instance):
    """Return a cached :class:`lib389.dseldif.DSEldif` for local ``dse.ldif`` reads in lint code."""
    if not getattr(instance, 'isLocal', False):
        return None
    if getattr(instance, '_lib389_dse_lint_cache', None) is not None:
        return instance._lib389_dse_lint_cache
    if getattr(instance, '_lib389_dse_lint_failed', False):
        return None
    try:
        from lib389.dseldif import DSEldif
        dse = DSEldif(instance)
    except (OSError, IOError, EnvironmentError):
        instance._lib389_dse_lint_failed = True
        return None
    instance._lib389_dse_lint_cache = dse
    return dse


def lint_get_attr_val_utf8(obj, key):
    """Like :meth:`DSLdapObject.get_attr_val_utf8`, but read ``dse.ldif`` when the server is down."""
    if obj._instance.state == DIRSRV_STATE_ONLINE:
        return obj.get_attr_val_utf8(key)
    dse = _cached_dse_ldif_for_lint(obj._instance)
    if dse is None:
        raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")

    vals = dse.get(ensure_str(obj._dn).lower(), ensure_str(key), single=False, lower=True)
    if vals is None:
        return None
    if len(vals) == 0:
        return ''
    return ensure_str(vals[0])


def lint_get_attr_val_utf8_l(obj, key):
    from lib389.utils import ensure_str
    v = lint_get_attr_val_utf8(obj, key)
    if v is None:
        return None
    return ensure_str(v).lower()


def lint_get_attr_vals_utf8(obj, key):
    """Like :meth:`DSLdapObject.get_attr_vals_utf8` for one attribute, with offline ``dse.ldif`` support."""
    if obj._instance.state == DIRSRV_STATE_ONLINE:
        return obj.get_attr_vals_utf8(key)
    dse = _cached_dse_ldif_for_lint(obj._instance)
    if dse is None:
        raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
    vals = dse.get(ensure_str(obj._dn).lower(), ensure_str(key), single=False, lower=True)
    if vals is None:
        return []
    return ensure_list_str(vals)


def lint_get_attr_vals_utf8_l(obj, key):
    from lib389.utils import ensure_str
    return [ensure_str(x).lower() for x in lint_get_attr_vals_utf8(obj, key)]


def lint_get_attr_val(obj, key):
    """Like :meth:`DSLdapObject.get_attr_val` (bytes), with offline ``dse.ldif`` support."""
    if obj._instance.state == DIRSRV_STATE_ONLINE:
        return obj.get_attr_val(key)
    s = lint_get_attr_val_utf8(obj, key)
    if s is None:
        return None
    if s == '':
        return ''
    return ensure_bytes(s)


def lint_get_attr_val_int(obj, key):
    if obj._instance.state == DIRSRV_STATE_ONLINE:
        return obj.get_attr_val_int(key)
    v = lint_get_attr_val_utf8(obj, key)
    if v == '' or v is None:
        return None
    return ensure_int(v)


def lint_plugin_enabled(plugin):
    """Whether ``nsslapd-pluginEnabled`` is ``on``; works offline via ``dse.ldif``."""
    return lint_get_attr_val_utf8(plugin, 'nsslapd-pluginEnabled') == 'on'



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import shutil
import subprocess
from lib389.nss_ssl import NssSsl
from lib389.utils import (
    assert_c,
    ensure_list_str,
    ensure_str,
    selinux_clean_files_label,
    selinux_clean_ports_label,
    selinux_label_file,
    selinux_label_port,
)


######################## WARNING #############################
# DO NOT CHANGE THIS FILE OR ITS CONTENTS WITHOUT READING
# ALL OF THE COMMENTS FIRST. THERE ARE VERY DELICATE
# AND DETAILED INTERACTIONS OF COMPONENTS IN THIS FILE.
#
# IF IN DOUBT CONTACT WILLIAM BROWN <william@blackhats.net.au>


def remove_ds_instance(dirsrv, force=False):
    """
    This will delete the instance as it is define. This must be a local instance. This is
    designed to raise exceptions quickly and often if *any* error is hit. However, this can
    be run repeatedly, and only when the instance is truely removed, will this program fail
    to run further.

    :param dirsrv: A directory server instance
    :type dirsrv: DirSrv
    :param force: A psychological aid, for people who think force means do something, harder. Does
        literally nothing in this program because state machines are a thing.
    :type force: bool
    """
    _log = dirsrv.log.getChild('remove_ds')
    _log.debug("Removing instance %s" % dirsrv.serverid)

    # Copy all the paths we are about to tamper with
    remove_paths = {}
    remove_paths['backup_dir'] = dirsrv.ds_paths.backup_dir
    remove_paths['cert_dir'] = dirsrv.ds_paths.cert_dir
    remove_paths['config_dir'] = dirsrv.ds_paths.config_dir
    remove_paths['db_dir'] = dirsrv.ds_paths.db_dir
    remove_paths['db_home_dir'] = dirsrv.ds_paths.db_home_dir
    remove_paths['db_dir_parent'] = dirsrv.ds_paths.db_dir + "/../"
    ### WARNING: The changelogdb isn't removed. we assume it's in:
    # db_dir ../changelogdb. So remove that too!
    # abspath will resolve the ".." down.
    remove_paths['changelogdb_dir'] = dirsrv.get_changelog_dir()
    remove_paths['ldif_dir'] = dirsrv.ds_paths.ldif_dir
    remove_paths['lock_dir'] = dirsrv.ds_paths.lock_dir
    remove_paths['log_dir'] = dirsrv.ds_paths.log_dir
    remove_paths['inst_dir'] = dirsrv.ds_paths.inst_dir
    remove_paths['etc_sysconfig'] = "%s/sysconfig/dirsrv-%s" % (dirsrv.ds_paths.sysconf_dir, dirsrv.serverid)
    remove_paths['ldapi'] = dirsrv.ds_paths.ldapi

    tmpfiles_d_path = dirsrv.ds_paths.tmpfiles_d + "/dirsrv-" + dirsrv.serverid + ".conf"

    # These are handled in a special way.
    dse_ldif_path = os.path.join(dirsrv.ds_paths.config_dir, 'dse.ldif')

    # Check the marker exists. If it *does not* warn about this, and say that to
    # force removal you should touch this file.

    _log.debug("Checking for instance marker at %s" % dse_ldif_path)
    if not os.path.exists(dse_ldif_path):
        _log.info("Instance configuration not found, no action will be taken")
        _log.info("If you want us to cleanup anyway, recreate '%s'" % dse_ldif_path)
        return
    _log.debug("Found instance marker at %s! Proceeding to remove ..." % dse_ldif_path)

    # Stop the instance (if running) and now we know it really does exist
    # and hopefully have permission to access it ...
    if dirsrv.status():
        _log.debug("Stopping instance %s" % dirsrv.serverid)
        dirsrv.stop()

    _log.debug("Found instance marker at %s! Proceeding to remove ..." % dse_ldif_path)

    ### ANY NEW REMOVAL ACTION MUST BE BELOW THIS LINE!!!

    # Remove these paths:
    # for path in ('backup_dir', 'cert_dir', 'config_dir', 'db_dir',
    #             'ldif_dir', 'lock_dir', 'log_dir', 'run_dir'):
    for path_k in remove_paths:
        _log.debug("Removing %s" % remove_paths[path_k])
        shutil.rmtree(remove_paths[path_k], ignore_errors=True)

    # Remove parent (/var/lib/dirsrv/slapd-INST)
    shutil.rmtree(remove_paths['db_dir'].replace('db', ''), ignore_errors=True)

    # Remove /run/slapd-isntance
    try:
        os.remove(f'/run/slapd-{dirsrv.serverid}.socket')
    except OSError as e:
        _log.debug("Failed to remove socket file: " + str(e))

    # We can not assume we have systemd ...
    if dirsrv.ds_paths.with_systemd:
        # Remove the systemd symlink
        _log.debug("Removing the systemd symlink")

        result = subprocess.run(["systemctl", "disable", "dirsrv@{}".format(dirsrv.serverid)],
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        args = ' '.join(ensure_list_str(result.args))
        stdout = ensure_str(result.stdout)
        stderr = ensure_str(result.stderr)
        _log.debug(f"CMD: {args} ; STDOUT: {stdout} ; STDERR: {stderr}")

        _log.debug("Removing %s" % tmpfiles_d_path)
        try:
            os.remove(tmpfiles_d_path)
        except OSError as e:
            _log.debug("Failed to remove tmpfile: " + str(e))

    # Nor can we assume we have SELinux.
    if dirsrv.ds_paths.with_selinux:
        # Remove selinux port label
        _log.debug("Removing the port labels")
        selinux_label_port(dirsrv.port, remove_label=True)

        # This is a compatability with ancient installs, all modern install have tls port
        if dirsrv.sslport is not None:
            selinux_label_port(dirsrv.sslport, remove_label=True)

    # If this was the last instance, remove the ssca instance
    # and all ds related SELinux customizations
    insts = dirsrv.list(all=True)
    if len(insts) == 0:
        ssca = NssSsl(dbpath=dirsrv.get_ssca_dir())
        ssca.remove_db()
        selinux_clean_ports_label()
        if dirsrv.ds_paths.prefix != '/usr':
            selinux_clean_files_label(all=True)
    else:
        if dirsrv.ds_paths.prefix != '/usr':
            selinux_clean_files_label()

    ### ANY NEW REMOVAL ACTIONS MUST BE ABOVE THIS LINE!!!

    # Finally means FINALLY, the last thing, the LAST LAST thing. By doing this absolutely
    # last, it means that we can have any failure above, and continue to re-run until resolved
    # because this instance marker (dse.ldif) continues to exist!
    # Move the config_dir to config_dir.removed
    config_dir = dirsrv.ds_paths.config_dir
    config_dir_rm = "{}.removed".format(config_dir)

    if os.path.exists(config_dir_rm):
        _log.debug("Removing previously existed %s" % config_dir_rm)
        shutil.rmtree(config_dir_rm)

    assert_c(not os.path.exists(config_dir_rm))
    selinux_label_file(config_dir_rm, None)

    # That's it, everything before this MUST have suceeded, so now we can move the
    # config dir (containing dse.ldif, the marker) out of the way.
    _log.debug("Moving %s to %s" % (config_dir, config_dir_rm))
    try:
        shutil.move(config_dir, config_dir_rm)
    except FileNotFoundError:
        pass

    # DO NOT PUT ANY CODE BELOW THIS COMMENT BECAUSE THAT WOULD VIOLATE THE ASSERTIONS OF THE
    # ABOVE CODE.

    # Done!
    _log.debug("Complete")



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016-2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest

from lib389.tests.cli import topology as default_topology
from lib389.cli_base import LogCapture, FakeArgs
from lib389.plugins import MemberOfPlugin
from lib389.cli_conf.plugins import memberof as memberof_cli


@pytest.fixture(scope="module")
def topology(request):
    topology = default_topology(request)

    plugin = MemberOfPlugin(topology.standalone)
    if not plugin.exists():
        plugin.create()

    # At the moment memberof plugin needs to be enabled in order to perform
    # syntax checking. Additionally, we have to restart the server in order
    # for the action of enabling the plugin to take effect.
    plugin.enable()
    topology.standalone.restart()
    topology.logcap.flush()

    return topology


def test_set_attr_with_legal_value(topology):
    args = FakeArgs()

    args.value = "memberOf"
    memberof_cli.manage_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfAttr set to")
    topology.logcap.flush()

    args.value = None
    memberof_cli.manage_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": memberOf")
    topology.logcap.flush()

def test_set_attr_with_illegal_value(topology):
    args = FakeArgs()

    args.value = "whatever"
    memberof_cli.manage_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Failed to set")
    topology.logcap.flush()

def test_set_groupattr_with_legal_value(topology):
    args = FakeArgs()

    args.value = "uniquemember"
    memberof_cli.add_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully added memberOfGroupAttr value")
    topology.logcap.flush()

    memberof_cli.display_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": uniquemember")
    topology.logcap.flush()

def test_set_groupattr_with_value_that_already_exists(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "uniquemember" in plugin.get_groupattr():
        plugin.add_groupattr("uniquemember")

    args = FakeArgs()

    args.value = "uniquemember"
    memberof_cli.add_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("already exists")
    topology.logcap.flush()

def test_set_groupattr_with_illegal_value(topology):
    args = FakeArgs()

    args.value = "whatever"
    memberof_cli.add_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Illegal value")
    topology.logcap.flush()

def test_remove_groupattr_with_value_that_exists(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "uniquemember" in plugin.get_groupattr():
        plugin.add_groupattr("uniquemember")

    args = FakeArgs()

    args.value = "uniquemember"
    memberof_cli.remove_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully removed memberOfGroupAttr value")
    topology.logcap.flush()

    memberof_cli.display_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert not topology.logcap.contains(": uniquemember")
    topology.logcap.flush()

def test_remove_groupattr_with_value_that_doesnt_exist(topology):
    args = FakeArgs()

    args.value = "whatever"
    memberof_cli.remove_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('No value "{0}" found'.format(args.value))
    topology.logcap.flush()

def test_try_remove_all_groupattr_values(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # make sure "member" value exists and it is the only one
    assert "member" in plugin.get_groupattr() # exists from default
    assert len(plugin.get_groupattr()) == 1

    args = FakeArgs()

    args.value = "member"
    memberof_cli.remove_groupattr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Error: Failed to delete. memberOfGroupAttr is required.")
    topology.logcap.flush()

def test_get_allbackends_when_not_set(topology):
    args = FakeArgs()

    memberof_cli.display_allbackends(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfAllBackends is not set")
    topology.logcap.flush()

def test_enable_allbackends(topology):
    args = FakeArgs()

    memberof_cli.enable_allbackends(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("enabled successfully")
    topology.logcap.flush()

    memberof_cli.display_allbackends(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": on")
    topology.logcap.flush()

def test_disable_all_backends(topology):
    args = FakeArgs()

    memberof_cli.disable_allbackends(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("disabled successfully")
    topology.logcap.flush()

    memberof_cli.display_allbackends(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": off")
    topology.logcap.flush()

def test_get_skipnested_when_not_set(topology):
    args = FakeArgs()

    memberof_cli.display_skipnested(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfSkipNested is not set")
    topology.logcap.flush()

def test_set_skipnested(topology):
    args = FakeArgs()

    memberof_cli.enable_skipnested(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("set successfully")
    topology.logcap.flush()

    memberof_cli.display_skipnested(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": on")
    topology.logcap.flush()

def test_unset_skipnested(topology):
    args = FakeArgs()

    memberof_cli.disable_skipnested(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("unset successfully")
    topology.logcap.flush()

    memberof_cli.display_skipnested(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": off")
    topology.logcap.flush()

def test_set_autoaddoc(topology):
    args = FakeArgs()

    # argparse makes sure that only choices 'nsmemberof', 'inetuser', 'inetadmin', 'del' are allowed
    args.value = "nsmemberof"
    memberof_cli.manage_autoaddoc(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfAutoAddOc set to")
    topology.logcap.flush()

    args.value = None
    memberof_cli.manage_autoaddoc(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": nsMemberOf")
    topology.logcap.flush()

def test_remove_autoaddoc(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not plugin.get_autoaddoc():
        plugin.set_autoaddoc("nsmemberof")

    args = FakeArgs()

    # argparse makes sure that only choices 'nsmemberof', 'inetuser', 'inetadmin', 'del' are allowed
    args.value = "del"
    memberof_cli.manage_autoaddoc(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfAutoAddOc attribute deleted")
    topology.logcap.flush()

    args.value = None
    memberof_cli.manage_autoaddoc(topology.standalone, None, topology.logcap.log, args)
    assert not topology.logcap.contains(": nsMemberOf")
    topology.logcap.flush()

def test_remove_autoaddoc_when_not_set(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if plugin.get_autoaddoc():
        plugin.remove_autoaddoc()

    args = FakeArgs()

    args.value = "del"
    memberof_cli.manage_autoaddoc(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfAutoAddOc was not set")
    topology.logcap.flush()

def test_get_autoaddoc_when_not_set(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if plugin.get_autoaddoc():
        plugin.remove_autoaddoc()

    args = FakeArgs()

    args.value = None
    memberof_cli.manage_autoaddoc(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfAutoAddOc is not set")
    topology.logcap.flush()

def test_get_entryscope_when_not_set(topology):
    args = FakeArgs()

    args.value = None
    memberof_cli.display_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfEntryScope is not set")
    topology.logcap.flush()

def test_add_entryscope_with_legal_value(topology):
    args = FakeArgs()

    args.value = "dc=example,dc=com"
    memberof_cli.add_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully added memberOfEntryScope value")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": dc=example,dc=com")
    topology.logcap.flush()

    args.value = "a=b"
    memberof_cli.add_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully added memberOfEntryScope value")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": dc=example,dc=com")
    assert topology.logcap.contains(": a=b")
    topology.logcap.flush()

def test_add_entryscope_with_illegal_value(topology):
    args = FakeArgs()

    args.value = "whatever"
    memberof_cli.add_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Error: Invalid DN")
    topology.logcap.flush()

def test_add_entryscope_with_existing_value(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "dc=example,dc=com" in plugin.get_entryscope():
        plugin.add_entryscope("dc=example,dc=com")

    args = FakeArgs()

    args.value = "dc=example,dc=com"
    memberof_cli.add_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('Value "{}" already exists'.format(args.value))
    topology.logcap.flush()

def test_remove_entryscope_with_existing_value(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "a=b" in plugin.get_entryscope():
        plugin.add_entryscope("a=b")
    if not "dc=example,dc=com" in plugin.get_entryscope():
        plugin.add_entryscope("dc=example,dc=com")

    args = FakeArgs()

    args.value = "a=b"
    memberof_cli.remove_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully removed memberOfEntryScope value")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": dc=example,dc=com")
    assert not topology.logcap.contains(": a=b")
    topology.logcap.flush()

def test_remove_entryscope_with_non_existing_value(topology):
    args = FakeArgs()

    args.value = "whatever"
    memberof_cli.remove_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('No value "{0}" found'.format(args.value))
    topology.logcap.flush()

def test_remove_all_entryscope(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "dc=example,dc=com" in plugin.get_entryscope():
        plugin.add_entryscope("dc=example,dc=com")
    if not "a=b" in plugin.get_entryscope():
        plugin.add_entryscope("a=b")

    args = FakeArgs()

    args.value = None
    memberof_cli.display_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": a=b")
    assert topology.logcap.contains(": dc=example,dc=com")
    topology.logcap.flush()

    args.value = None
    memberof_cli.remove_all_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully removed all memberOfEntryScope values")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_scope(topology.standalone, None, topology.logcap.log, args)
    assert not topology.logcap.contains(": a=b")
    assert not topology.logcap.contains(": dc=example,dc=com")
    topology.logcap.flush()

def test_get_excludescope_when_not_set(topology):
    args = FakeArgs()

    args.value = None
    memberof_cli.display_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("memberOfEntryScopeExcludeSubtree is not set")
    topology.logcap.flush()

def test_add_excludescope_with_legal_value(topology):
    args = FakeArgs()

    args.value = "ou=people,dc=example,dc=com"
    memberof_cli.add_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully added memberOfEntryScopeExcludeSubtree value")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": ou=people,dc=example,dc=com")
    topology.logcap.flush()

    args.value = "a=b"
    memberof_cli.add_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully added memberOfEntryScopeExcludeSubtree value")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": ou=people,dc=example,dc=com")
    assert topology.logcap.contains(": a=b")
    topology.logcap.flush()

def test_add_excludescope_with_illegal_value(topology):
    args = FakeArgs()

    args.value = "whatever"
    memberof_cli.add_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Error: Invalid DN")
    topology.logcap.flush()

def test_add_excludescope_with_existing_value(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "ou=people,dc=example,dc=com" in plugin.get_excludescope():
        plugin.add_excludescope("ou=people,dc=example,dc=com")

    args = FakeArgs()

    args.value = "ou=people,dc=example,dc=com"
    memberof_cli.add_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('Value "{}" already exists'.format(args.value))
    topology.logcap.flush()

def test_remove_excludescope_with_existing_value(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "a=b" in plugin.get_excludescope():
        plugin.add_excludescope("a=b")
    if not "ou=people,dc=example,dc=com" in plugin.get_excludescope():
        plugin.add_excludescope("ou=people,dc=example,dc=com")

    args = FakeArgs()

    args.value = "a=b"
    memberof_cli.remove_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully removed memberOfEntryScopeExcludeSubtree value")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": ou=people,dc=example,dc=com")
    assert not topology.logcap.contains(": a=b")
    topology.logcap.flush()

def test_remove_excludescope_with_non_existing_value(topology):
    args = FakeArgs()

    args.value = "whatever"
    memberof_cli.remove_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('No value "{0}" found'.format(args.value))
    topology.logcap.flush()

def test_remove_all_excludescope(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "a=b" in plugin.get_excludescope():
        plugin.add_excludescope("a=b")
    if not "ou=people,dc=example,dc=com" in plugin.get_excludescope():
        plugin.add_excludescope("ou=people,dc=example,dc=com")

    args = FakeArgs()

    args.value = None
    memberof_cli.display_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": a=b")
    assert topology.logcap.contains(": ou=people,dc=example,dc=com")
    topology.logcap.flush()

    args.value = None
    memberof_cli.remove_all_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully removed all memberOfEntryScopeExcludeSubtree values")
    topology.logcap.flush()

    args.value = None
    memberof_cli.display_excludescope(topology.standalone, None, topology.logcap.log, args)
    assert not topology.logcap.contains(": a=b")
    assert not topology.logcap.contains(": ou=people,dc=example,dc=com")
    topology.logcap.flush()

def test_add_entryscope_with_value_that_exists_in_excludescope(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "dc=example,dc=com" in plugin.get_entryscope():
        plugin.add_entryscope("dc=example,dc=com")
    if not "ou=people,dc=example,dc=com" in plugin.get_excludescope():
        plugin.add_excludescope("ou=people,dc=example,dc=com")

    args = FakeArgs()

    args.value = "ou=people,dc=example,dc=com"
    memberof_cli.add_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("is also listed as an exclude suffix")
    topology.logcap.flush()

def test_add_excludescope_with_value_that_exists_in_entryscope(topology):
    plugin = MemberOfPlugin(topology.standalone)
    # setup test
    if not "dc=example,dc=com" in plugin.get_entryscope():
        plugin.add_entryscope("dc=example,dc=com")
    if not "ou=people,dc=example,dc=com" in plugin.get_excludescope():
        plugin.add_excludescope("ou=people,dc=example,dc=com")

    args = FakeArgs()

    args.value = "ou=people,dc=example,dc=com"
    memberof_cli.add_scope(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("is also listed as an exclude suffix")
    topology.logcap.flush()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import collections
import ldap
import copy
import os.path
from lib389.utils import *
from lib389._entry import Entry
from lib389.idm.nscontainer import nsContainers
from lib389._mapped_object import DSLdapObjects, DSLdapObject
from lib389._constants import DEFAULT_SUFFIX

class Rewriter(DSLdapObject):
    """A single instance of a rewriter entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(Rewriter, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['nsslapd-libpath']
        self._create_objectclasses = ['top', 'rewriterEntry']


class Rewriters(DSLdapObjects):
    """A DSLdapObjects entity which represents rewriter entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Base DN for all account entries below
    :type basedn: str
    """

    def __init__(self, instance, basedn=None):
        super(Rewriters, self).__init__(instance=instance)
        self._objectclasses = ['top', 'rewriterEntry']
        self._filterattrs = ['cn', 'nsslapd-libpath']
        self._childobject = Rewriter
        self._basedn = 'cn=rewriters,cn=config'
        self._list_attrlist = ['dn', 'nsslapd-libpath']


class AdRewriter(Rewriter):
    """An instance of AD rewriter entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn="cn=adrewriter,cn=rewriter,cn=config"):
        super(AdRewriter, self).__init__(instance, dn)
        self._configuration_dn = None
        self._schema_dn = None

    def create_containers(self, suffix):
        conf_conts = nsContainers(self._instance, suffix)
        conf_cont = conf_conts.ensure_state(properties={'cn': 'Configuration'})
        schema_conts = nsContainers(self._instance, conf_cont.dn)
        schema_cont = schema_conts.ensure_state(properties={'cn': 'Schema'})
        self._configuration_dn = conf_cont.dn
        self._schema_dn = schema_cont.dn

    def get_schema_dn(self):
        return self._schema_dn


class AdRewriters(Rewriters):
    """A DSLdapObjects entity which represents AD rewriter entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Base DN for all account entries below
    :type basedn: str
    """

    def __init__(self, instance, basedn=None):
        super(AdRewriters, self).__init__(instance=instance)
        self._childobject = AdRewriter





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from .config import baseconfig
from .sample import sampleentries
from lib389.idm.organizationalunit import OrganizationalUnits
from lib389.idm.group import Groups
from lib389.idm.posixgroup import PosixGroups
from lib389.idm.user import nsUserAccounts
from lib389.idm.nscontainer import nsHiddenContainers


class c001004000_sample_entries(sampleentries):
    def __init__(self, instance, basedn):
        super(c001004000_sample_entries, self).__init__(instance, basedn)
        self.description = """Apply sample entries matching the 1.4.0 sample data and access controls"""
        self.version = "c001004000"

    # All checks done, apply!
    def _apply(self):
        self._configure_base()

        # Create the 389 service container
        # This could also move to be part of core later ....
        hidden_containers = nsHiddenContainers(self._instance, self._basedn)
        hidden_containers.create(properties={
            'cn': '389_ds_system'
            })

        # Create our ous.
        ous = OrganizationalUnits(self._instance, self._basedn)
        ous.create(properties = {
            'ou': 'groups',
            'aci': [
                # Allow anon partial read
                '(targetattr="cn || member || gidNumber || nsUniqueId || description || objectClass")(targetfilter="(objectClass=groupOfNames)")(version 3.0; acl "Enable anyone group read"; allow (read, search, compare)(userdn="ldap:///anyone");)',
                # Allow group_modify to modify but not create groups
                '(targetattr="member")(targetfilter="(objectClass=groupOfNames)")(version 3.0; acl "Enable group_modify to alter members"; allow (write)(groupdn="ldap:///cn=group_modify,ou=permissions,{BASEDN}");)'.format(BASEDN=self._basedn),
                # Allow group_admin to fully manage groups (posix or not).
                '(targetattr="cn || member || gidNumber || description || objectClass")(targetfilter="(objectClass=groupOfNames)")(version 3.0; acl "Enable group_admin to manage groups"; allow (write, add, delete)(groupdn="ldap:///cn=group_admin,ou=permissions,{BASEDN}");)'.format(BASEDN=self._basedn),
            ]
        })

        ous.create(properties = {
            'ou': 'people',
            'aci': [
                # allow anon partial read.
                '(targetattr="objectClass || description || nsUniqueId || uid || displayName || loginShell || uidNumber || gidNumber || gecos || homeDirectory || cn || memberOf || mail || nsSshPublicKey || nsAccountLock || userCertificate")(targetfilter="(objectClass=posixaccount)")(version 3.0; acl "Enable anyone user read"; allow (read, search, compare)(userdn="ldap:///anyone");)',
                # allow self partial mod
                '(targetattr="displayName || nsSshPublicKey")(version 3.0; acl "Enable self partial modify"; allow (write)(userdn="ldap:///self");)',
                # Allow self full read
                '(targetattr="legalName || telephoneNumber || mobile")(targetfilter="(objectClass=nsPerson)")(version 3.0; acl "Enable self legalname read"; allow (read, search, compare)(userdn="ldap:///self");)',
                # Allow reading legal name
                '(targetattr="legalName || telephoneNumber")(targetfilter="(objectClass=nsPerson)")(version 3.0; acl "Enable user legalname read"; allow (read, search, compare)(groupdn="ldap:///cn=user_private_read,ou=permissions,{BASEDN}");)'.format(BASEDN=self._basedn),
                # These below need READ so they can read userPassword and legalName
                # Allow user admin create mod
                '(targetattr="uid || description || displayName || loginShell || uidNumber || gidNumber || gecos || homeDirectory || cn || memberOf || mail || legalName || telephoneNumber || mobile")(targetfilter="(&(objectClass=nsPerson)(objectClass=nsAccount))")(version 3.0; acl "Enable user admin create"; allow (write, add, delete, read)(groupdn="ldap:///cn=user_admin,ou=permissions,{BASEDN}");)'.format(BASEDN=self._basedn),
                # Allow user mod mod only
                '(targetattr="uid || description || displayName || loginShell || uidNumber || gidNumber || gecos || homeDirectory || cn || memberOf || mail || legalName || telephoneNumber || mobile")(targetfilter="(&(objectClass=nsPerson)(objectClass=nsAccount))")(version 3.0; acl "Enable user modify to change users"; allow (write, read)(groupdn="ldap:///cn=user_modify,ou=permissions,{BASEDN}");)'.format(BASEDN=self._basedn),
                # Allow user_pw_admin to nsaccountlock and password
                '(targetattr="userPassword || nsAccountLock || userCertificate || nsSshPublicKey")(targetfilter="(objectClass=nsAccount)")(version 3.0; acl "Enable user password reset"; allow (write, read)(groupdn="ldap:///cn=user_passwd_reset,ou=permissions,{BASEDN}");)'.format(BASEDN=self._basedn),
            ]
        })

        ous.create(properties = {
            'ou': 'permissions',
        })

        ous.create(properties = {
            'ou': 'services',
            'aci': [
                # Minimal service read
                '(targetattr="objectClass || description || nsUniqueId || cn || memberOf || nsAccountLock ")(targetfilter="(objectClass=netscapeServer)")(version 3.0; acl "Enable anyone service account read"; allow (read, search, compare)(userdn="ldap:///anyone");)',
            ]
        })

        # Create the demo user
        users = nsUserAccounts(self._instance, self._basedn)
        users.create(properties={
            'uid': 'demo_user',
            'cn': 'Demo User',
            'displayName': 'Demo User',
            'legalName': 'Demo User Name',
            'uidNumber': '99998',
            'gidNumber': '99998',
            'homeDirectory': '/var/empty',
            'loginShell': '/bin/false',
            'nsAccountlock': 'true'
        })

        # Create the demo group
        groups = PosixGroups(self._instance, self._basedn)
        groups.create(properties={
            'cn' : 'demo_group',
            'gidNumber': '99999'
        })

        # Create the permission groups required for the acis
        permissions = Groups(self._instance, self._basedn, rdn='ou=permissions')
        permissions.create(properties={
            'cn': 'group_admin',
        })
        permissions.create(properties={
            'cn': 'group_modify',
        })
        permissions.create(properties={
            'cn': 'user_admin',
        })
        permissions.create(properties={
            'cn': 'user_modify',
        })
        permissions.create(properties={
            'cn': 'user_passwd_reset',
        })
        permissions.create(properties={
            'cn': 'user_private_read',
        })


class c001004000(baseconfig):
    def __init__(self, instance):
        super(c001004000, self).__init__(instance)
        self._operations = [
            # For now this is an empty place holder - likely this
            # will become part of core server.
        ]



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import socket
from lib389.plugins import RootDNAccessControlPlugin
from lib389.utils import is_valid_hostname, is_valid_ip
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit
from lib389.cli_base import CustomHelpFormatter

arg_to_attr = {
    'allow_host': 'rootdn-allow-host',
    'deny_host': 'rootdn-deny-host',
    'allow_ip': 'rootdn-allow-ip',
    'deny_ip': 'rootdn-deny-ip',
    'open_time': 'rootdn-open-time',
    'close_time': 'rootdn-close-time',
    'days_allowed': 'rootdn-days-allowed'
}


def validate_args(args):
    # validate the args
    if args.close_time is not None:
        try:
            int(args.close_time)
        except:
            raise ValueError("The close time must be a 4 digit number: HHMM")
        if len(args.close_time) != 4:
            raise ValueError("The close time must be a 4 digit number: HHMM")
        hour = int(args.close_time[:2])
        if hour < 0 or hour > 23:
            raise ValueError(f"The hour portion of the time is invalid: {hour}  Must be between 0 and 23")
        min = int(args.close_time[-2:])
        if min < 0 or min > 59:
            raise ValueError(f"The minute portion of the time is invalid: {min}  Must be between 1 and 59")

    if args.open_time is not None:
        try:
            int(args.open_time)
        except:
            raise ValueError("The open time must be a 4 digit number: HHMM")
        if len(args.open_time) != 4:
            raise ValueError("The open time must be a 4 digit number: HHMM")
        hour = int(args.open_time[:2])
        if hour < 0 or hour > 23:
            raise ValueError(f"The hour portion of the time is invalid: {hour}  Must be between 0 and 23")
        min = int(args.open_time[-2:])
        if min < 0 or min > 59:
            raise ValueError(f"The minute portion of the time is invalid: {min}  Must be between 1 and 59")

    if args.days_allowed is not None and args.days_allowed != "delete":
        valid_days = ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun']
        choosen_days =  args.days_allowed.lower().replace(' ', '').split(',')
        for day in choosen_days:
            if day not in valid_days:
                raise ValueError(f"Invalid day entered ({day}), valid days are: Mon, Tue, Wed, Thu, Fri, Sat, Sun")

    if args.allow_ip is not None:
        for ip in args.allow_ip:
            if ip != "delete" and not is_valid_ip(ip):
                raise ValueError(f"Invalid IP address ({ip}) for '--allow-ip'")

    if args.deny_ip is not None and args.deny_ip != "delete":
        for ip in args.deny_ip:
            if ip != "delete" and not is_valid_ip(ip):
                raise ValueError(f"Invalid IP address ({ip}) for '--deny-ip'")

    if args.allow_host is not None:
        for hostname in args.allow_host:
            if hostname != "delete" and not is_valid_hostname(hostname):
                raise ValueError(f"Invalid hostname ({hostname}) for '--allow-host'")

    if args.deny_host is not None:
        for hostname in args.deny_host:
            if hostname != "delete" and not is_valid_hostname(hostname):
                raise ValueError(f"Invalid hostname ({hostname}) for '--deny-host'")


def rootdn_edit(inst, basedn, log, args):
    log = log.getChild('rootdn_edit')
    validate_args(args)
    plugin = RootDNAccessControlPlugin(inst)
    generic_object_edit(plugin, log, args, arg_to_attr)


def _add_parser_args(parser):
    parser.add_argument('--allow-host', nargs='+',
                        help='Sets what hosts, by fully-qualified domain name, the root user is allowed to use '
                             'to access Directory Server. Any hosts not listed are implicitly denied '
                             '(rootdn-allow-host)')
    parser.add_argument('--deny-host', nargs='+',
                        help='Sets what hosts, by fully-qualified domain name, the root user is not allowed to use '
                             'to access Directory Server. Any hosts not listed are implicitly allowed '
                             '(rootdn-deny-host). If a host address is listed in both the rootdn-allow-host and '
                             'rootdn-deny-host attributes, it is denied access.')
    parser.add_argument('--allow-ip', nargs='+',
                        help='Sets what IP addresses, either IPv4 or IPv6, for machines the root user is allowed '
                             'to use to access Directory Server. Any IP addresses not listed are implicitly '
                             'denied (rootdn-allow-ip)')
    parser.add_argument('--deny-ip', nargs='+',
                        help='Sets what IP addresses, either IPv4 or IPv6, for machines the root user is not allowed '
                             'to use to access Directory Server. Any IP addresses not listed are implicitly '
                             'allowed (rootdn-deny-ip). If an IP address is listed in both the rootdn-allow-ip and '
                             'rootdn-deny-ip attributes, it is denied access.')
    parser.add_argument('--open-time',
                        help='Sets part of a time period or range when the root user is allowed to access '
                             'Directory Server. This sets when the time-based access begins (rootdn-open-time)')
    parser.add_argument('--close-time',
                        help='Sets part of a time period or range when the root user is allowed to access '
                             'Directory Server. This sets when the time-based access ends (rootdn-close-time)')
    parser.add_argument('--days-allowed',
                        help='Sets a comma-separated list of what days the root user is allowed to use to access '
                             'Directory Server. Any days listed are implicitly denied (rootdn-days-allowed)')


def create_parser(subparsers):
    rootdnac_parser = subparsers.add_parser('root-dn', help='Manage and configure RootDN Access Control plugin', formatter_class=CustomHelpFormatter)
    subcommands = rootdnac_parser.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, RootDNAccessControlPlugin)

    edit = subcommands.add_parser('set', help='Edit the plugin settings', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=rootdn_edit)
    _add_parser_args(edit)





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.plugins import POSIXWinsyncPlugin
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit
from lib389.cli_base import CustomHelpFormatter

arg_to_attr = {
    'create_memberof_task': 'posixWinsyncCreateMemberOfTask',
    'lower_case_uid': 'posixWinsyncLowerCaseUID',
    'map_member_uid': 'posixWinsyncMapMemberUID',
    'map_nested_grouping': 'posixWinsyncMapNestedGrouping',
    'ms_sfu_schema': 'posixWinsyncMsSFUSchema'
}


def winsync_edit(inst, basedn, log, args):
    log = log.getChild('winsync_edit')
    plugin = POSIXWinsyncPlugin(inst)
    generic_object_edit(plugin, log, args, arg_to_attr)


def do_fixup(inst, basedn, log, args):
    plugin = POSIXWinsyncPlugin(inst)
    log.info('Attempting to add task entry...')
    if not plugin.status():
        log.error(f"'{plugin.rdn}' is disabled. Fix up task can't be executed")
        return
    fixup_task = plugin.fixup(args.DN, args.filter)
    fixup_task.wait(timeout=args.timeout)
    exitcode = fixup_task.get_exit_code()
    if exitcode != 0:
        if exitcode is None:
            raise ValueError(f'MemberUID task for {args.DN} has not completed. Please, check logs')
        else:
            raise ValueError(f'MemberUID task for {args.DN} has failed. Please, check logs')
    else:
        log.info('Successfully added task entry')


def _add_parser_args(parser):
    parser.add_argument('--create-memberof-task', choices=['true', 'false'], type=str.lower,
                        help='Sets whether to run the memberUID fix-up task immediately after a sync run in order '
                             'to update group memberships for synced users (posixWinsyncCreateMemberOfTask)')
    parser.add_argument('--lower-case-uid', choices=['true', 'false'], type=str.lower,
                        help='Sets whether to store (and, if necessary, convert) the UID value in the memberUID '
                             'attribute in lower case.(posixWinsyncLowerCaseUID)')
    parser.add_argument('--map-member-uid', choices=['true', 'false'], type=str.lower,
                        help='Sets whether to map the memberUID attribute in an Active Directory group to '
                             'the uniqueMember attribute in a Directory Server group (posixWinsyncMapMemberUID)')
    parser.add_argument('--map-nested-grouping', choices=['true', 'false'], type=str.lower,
                        help='Manages if nested groups are updated when memberUID attributes in '
                             'an Active Directory POSIX group change (posixWinsyncMapNestedGrouping)')
    parser.add_argument('--ms-sfu-schema', choices=['true', 'false'], type=str.lower,
                        help='Sets whether to the older Microsoft System Services for Unix 3.0 (msSFU30) '
                             'schema when syncing Posix attributes from Active Directory (posixWinsyncMsSFUSchema)')


def create_parser(subparsers):
    winsync = subparsers.add_parser('posix-winsync', help='Manage and configure the Posix Winsync API plugin', formatter_class=CustomHelpFormatter)
    subcommands = winsync.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, POSIXWinsyncPlugin)

    edit = subcommands.add_parser('set', help='Edit the plugin settings', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=winsync_edit)
    _add_parser_args(edit)

    fixup = subcommands.add_parser('fixup', help='Run the memberOf fix-up task to correct mismatched member and uniquemember values for synced users', formatter_class=CustomHelpFormatter)
    fixup.set_defaults(func=do_fixup)
    fixup.add_argument('DN', help="Set the base DN that contains entries to fix up")
    fixup.add_argument('-f', '--filter',
                       help='Filter for entries to fix up.\n If omitted, all entries with objectclass '
                            'inetuser/inetadmin/nsmemberof under the specified base will have '
                            'their memberOf attribute regenerated.')
    fixup.add_argument('--timeout', default=120, type=int,
                       help="Set a timeout to wait for the fixup task.  Default is 120 seconds")


# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

####################################
#
# Properties supported by the server
#
####################################

#
# Those WITH related attribute name
#
SER_HOST = 'hostname'
SER_PORT = 'ldap-port'
SER_LDAP_URL = 'ldapurl'
SER_SECURE_PORT = 'ldap-secureport'
SER_ROOT_DN = 'root-dn'
SER_ROOT_PW = 'root-pw'
SER_USER_ID = 'user-id'
SER_CREATION_SUFFIX = 'suffix'
SER_LDAPI_ENABLED = 'ldapi_enabled'
SER_LDAPI_SOCKET = 'ldapi_socket'
SER_LDAPI_AUTOBIND = 'ldapi_autobind'
SER_INST_SCRIPTS_ENABLED = 'InstScriptsEnabled'
SER_DB_LIB = 'db_lib'

SER_PROPNAME_TO_ATTRNAME = {SER_HOST: 'nsslapd-localhost',
                            SER_PORT: 'nsslapd-port',
                            SER_SECURE_PORT: 'nsslapd-securePort',
                            SER_ROOT_DN: 'nsslapd-rootdn',
                            # SER_ROOT_PW: 'nsslapd-rootpw', # We can't use this value ...
                            SER_USER_ID: 'nsslapd-localuser',
                            SER_CREATION_SUFFIX: 'nsslapd-defaultnamingcontext',
                            SER_LDAPI_ENABLED: 'nsslapd-ldapilisten',
                            SER_LDAPI_SOCKET: 'nsslapd-ldapifilepath',
                            SER_LDAPI_AUTOBIND: 'nsslapd-ldapiautobind',
                            SER_DB_LIB: 'nsslapd-backend-implement',
                            }
#
# Properties that are not associated with an attribute in cn=config entry
#
SER_SERVERID_PROP = 'server-id'
SER_GROUP_ID = 'group-id'
SER_DEPLOYED_DIR = 'deployed-dir'
SER_BACKUP_INST_DIR = 'inst-backupdir'
SER_STRICT_HOSTNAME_CHECKING = 'strict_hostname_checking'
SER_MDB_MAX_SIZE = 'mdb_max_size'

####################################
#
# Properties supported by the Mapping Tree entries
#
####################################

# Properties name
MT_STATE = 'state'
MT_BACKEND = 'backend-name'
MT_SUFFIX = 'suffix'
MT_REFERRAL = 'referral'
MT_CHAIN_PATH = 'chain-plugin-path'
MT_CHAIN_FCT = 'chain-plugin-fct'
MT_CHAIN_UPDATE = 'chain-update-policy'
MT_PARENT_SUFFIX = 'parent-suffix'

# Properties values
MT_STATE_VAL_DISABLED = 'disabled'
MT_STATE_VAL_BACKEND = 'backend'
MT_STATE_VAL_REFERRAL = 'referral'
MT_STATE_VAL_REFERRAL_ON_UPDATE = 'referral-on-update'
MT_STATE_VAL_CONTAINER = 'container'
MT_STATE_TO_VALUES = {MT_STATE_VAL_DISABLED: 0,
                      MT_STATE_VAL_BACKEND: 1,
                      MT_STATE_VAL_REFERRAL: 2,
                      MT_STATE_VAL_REFERRAL_ON_UPDATE: 3,
                      MT_STATE_VAL_CONTAINER: 4}

MT_CHAIN_UPDATE_VAL_ON_UPDATE = 'repl_chain_on_update'
MT_OBJECTCLASS_VALUE = 'nsMappingTree'

MT_PROPNAME_TO_VALUES = {MT_STATE: MT_STATE_TO_VALUES}
MT_PROPNAME_TO_ATTRNAME = {MT_STATE: 'nsslapd-state',
                           MT_BACKEND: 'nsslapd-backend',
                           MT_SUFFIX: 'cn',
                           MT_PARENT_SUFFIX: 'nsslapd-parent-suffix',
                           MT_REFERRAL: 'nsslapd-referral',
                           MT_CHAIN_PATH: 'nsslapd-distribution-plugin',
                           MT_CHAIN_FCT: 'nsslapd-distribution-funct',
                           MT_CHAIN_UPDATE: 'nsslapd-distribution-root-update'}

####################################
#
# Properties supported by the backend
#
####################################
BACKEND_SUFFIX = 'suffix'
BACKEND_NAME = 'name'
BACKEND_READONLY = 'read-only'
BACKEND_REQ_INDEX = 'require-index'
BACKEND_CACHE_ENTRIES = 'entry-cache-number'
BACKEND_CACHE_SIZE = 'entry-cache-size'
BACKEND_DNCACHE_SIZE = 'dn-cache-size'
BACKEND_DIRECTORY = 'directory'
BACKEND_DB_DEADLOCK = 'db-deadlock'
BACKEND_CHAIN_BIND_DN = 'chain-bind-dn'
BACKEND_CHAIN_BIND_PW = 'chain-bind-pw'
BACKEND_CHAIN_URLS = 'chain-urls'
BACKEND_STATS = 'stats'
BACKEND_SAMPLE_ENTRIES = 'sample_entries'
BACKEND_OBJECTCLASS_VALUE = 'nsBackendInstance'
BACKEND_REPL_ENABLED = 'enable_replication'
BACKEND_REPL_ROLE = 'replica_role'
BACKEND_REPL_ID = 'replica_id'
BACKEND_REPL_BINDDN = 'replica_binddn'
BACKEND_REPL_BINDPW = 'replica_bindpw'
BACKEND_REPL_BINDGROUP = 'replica_bindgroup'
BACKEND_REPL_CL_MAX_ENTRIES = "changelog_max_entries"
BACKEND_REPL_CL_MAX_AGE = "changelog_max_age"

# THIS NEEDS TO BE REMOVED. HACKS!!!!
BACKEND_PROPNAME_TO_ATTRNAME = {BACKEND_SUFFIX: 'nsslapd-suffix',
                                BACKEND_NAME: 'cn',
                                BACKEND_READONLY: 'nsslapd-readonly',
                                BACKEND_REQ_INDEX: 'nsslapd-require-index',
                                BACKEND_CACHE_ENTRIES: 'nsslapd-cachesize',
                                BACKEND_CACHE_SIZE: 'nsslapd-cachememsize',
                                BACKEND_DNCACHE_SIZE: 'nsslapd-dncachememsize',
                                BACKEND_DIRECTORY: 'nsslapd-directory',
                                BACKEND_DB_DEADLOCK:
                                    'nsslapd-db-deadlock-policy',
                                BACKEND_CHAIN_BIND_DN: 'nsmultiplexorbinddn',
                                BACKEND_CHAIN_BIND_PW:
                                    'nsmultiplexorcredentials',
                                BACKEND_CHAIN_URLS: 'nsfarmserverurl'}

####################################
#
# Properties of a replica
#
####################################


#
# General attributes (for both replica and agmt)
#
REPL_RUV = 'nsds50ruv'
REPL_ROOT = 'nsDS5ReplicaRoot'
REPL_PROTOCOL_TIMEOUT = 'nsds5ReplicaProtocolTimeout'
REPL_BINDDN = 'nsDS5ReplicaBindDN'
REPL_REAP_ACTIVE = 'nsds5ReplicaReapActive'

#
# Replica Entry Attributes
#
REPL_PURGE_DELAY = 'nsds5ReplicaPurgeDelay'
REPL_RETRY_MIN = 'nsds5ReplicaBackoffMin'
REPL_RETRY_MAX = 'nsds5ReplicaBackoffMax'
REPL_TYPE = 'nsDS5ReplicaType'
REPL_FLAGS = 'nsDS5Flags'
REPL_ID = 'nsDS5ReplicaId'
REPL_STATE = 'nsState'
REPL_NAME = 'nsDS5ReplicaName'
REPL_BIND_GROUP = 'nsds5replicabinddngroup'
REPL_BIND_GROUP_INTERVAL = 'nsds5replicabinddngroupcheckinterval'
REPL_REF = 'nsds5ReplicaReferral'
REPL_TOMBSTONE_PURGE_INTERVAL = 'nsds5ReplicaTombstonePurgeInterval'
REPL_CLEAN_RUV = 'nsds5ReplicaCleanRUV'
REPL_ABORT_RUV = 'nsds5ReplicaAbortCleanRUV'
REPL_COUNT_COUNT = 'nsds5ReplicaChangeCount'
REPL_PRECISE_PURGE = 'nsds5ReplicaPreciseTombstonePurging'
REPL_RELEASE_TIMEOUT = 'nsds5replicaReleaseTimeout'

# The values are from the REST API
REPLICA_SUFFIX = 'suffix'
REPLICA_PURGE_DELAY = 'ReplicaPurgeDelay'
REPLICA_ROOT = 'ReplicaRoot'
REPLICA_PROTOCOL_TIMEOUT = 'ReplicaProtocolTimeout'
REPLICA_BINDDN = 'ReplicaBindDN'
REPLICA_REAP_ACTIVE = 'ReplicaReapActive'
REPLICA_RETRY_MIN = 'ReplicaBackoffMin'
REPLICA_RETRY_MAX = 'ReplicaBackoffMax'
REPLICA_TYPE = 'ReplicaType'
REPLICA_FLAGS = 'ReplicaFlags'
REPLICA_ID = 'Replicaid'
REPLICA_STATE = 'ReplicaState'
REPLICA_NAME = 'ReplicaName'
REPLICA_BIND_GROUP = 'ReplicaBindDNGroup'
REPLICA_BIND_GROUP_INTERVAL = 'ReplicaBindDNGroupCheckInterval'
REPLICA_REFERRAL = 'ReplicaReferral'
REPLICA_TOMBSTONE_PURGE_INTERVAL = 'ReplicaTombstonePurgeInterval'
REPLICA_PURGE_INTERVAL = REPLICA_TOMBSTONE_PURGE_INTERVAL
REPLICA_CLEAN_RUV = 'ReplicaCleanRUV'
REPLICA_ABORT_RUV = 'ReplicaAbortCleanRUV'
REPLICA_COUNT_COUNT = 'ReplicaChangeCount'
REPLICA_PRECISE_PURGING = 'ReplicaPreciseTombstonePurging'
REPLICA_RELEASE_TIMEOUT = 'ReplicaReleaseTimeout'

REPLICA_PROPNAME_TO_ATTRNAME = {REPLICA_SUFFIX: REPL_ROOT,
                                REPLICA_ROOT: REPL_ROOT,
                                REPLICA_ID: REPL_ID,
                                REPLICA_TYPE: REPL_TYPE,
                                REPLICA_PURGE_INTERVAL:
                                    REPL_TOMBSTONE_PURGE_INTERVAL,
                                REPLICA_PURGE_DELAY: REPL_PURGE_DELAY,
                                REPLICA_REFERRAL: REPL_REF,
                                REPLICA_PROTOCOL_TIMEOUT:
                                    REPL_PROTOCOL_TIMEOUT,
                                REPLICA_BINDDN: REPL_BINDDN,
                                REPLICA_REAP_ACTIVE: REPL_REAP_ACTIVE,
                                REPLICA_RETRY_MIN: REPL_RETRY_MIN,
                                REPLICA_RETRY_MAX: REPL_RETRY_MAX,
                                REPLICA_FLAGS: REPL_FLAGS,
                                REPLICA_STATE: REPL_STATE,
                                REPLICA_NAME: REPL_NAME,
                                REPLICA_BIND_GROUP: REPL_BIND_GROUP,
                                REPLICA_BIND_GROUP_INTERVAL:
                                    REPL_BIND_GROUP_INTERVAL,
                                REPLICA_TOMBSTONE_PURGE_INTERVAL:
                                    REPL_TOMBSTONE_PURGE_INTERVAL,
                                REPLICA_CLEAN_RUV: REPL_CLEAN_RUV,
                                REPLICA_ABORT_RUV: REPL_ABORT_RUV,
                                REPLICA_COUNT_COUNT: REPL_COUNT_COUNT,
                                REPLICA_PRECISE_PURGING: REPL_PRECISE_PURGE,
                                REPLICA_RELEASE_TIMEOUT: REPL_RELEASE_TIMEOUT}

####################################
#
# Properties of a changelog
#
####################################


#
# Changelog Attributes
#
CL_DIR = 'nsslapd-changelogdir'
CL_MAX_ENTRIES = 'nsslapd-changelogmaxentries'
CL_MAXAGE = 'nsslapd-changelogmaxage'
CL_COMPACT_INTERVAL = 'nsslapd-changelogcompactdb-interval'
CL_TRIM_INTERVAL = 'nsslapd-changelogtrim-interval'
CL_CONCURRENT_WRITES = 'nsslapd-changelogmaxconcurrentwrites'
CL_ENCRYPT_ALG = 'nsslapd-encryptionalgorithm'
CL_SYM_KEY = 'nsSymmetricKey'

# REST names
CHANGELOG_NAME = 'cl-name'
CHANGELOG_DIR = 'ChangelogDir'
CHANGELOG_MAXAGE = 'ChangelogMaxAge'
CHANGELOG_MAXENTRIES = 'ChangelogMaxEntries'
CHANGELOG_TRIM_INTERVAL = 'ChangelogTrimInterval'
CHANGELOG_COMPACT_INTV = 'ChangelogCompactDBInterval'
CHANGELOG_CONCURRENT_WRITES = 'ChangelogMaxConcurrentWrites'
CHANGELOG_ENCRYPT_ALG = 'ChangelogEncryptionAlgorithm'
CHANGELOG_SYM_KEY = 'ChangelogSymmetricKey'

CL_DIR = 'nsslapd-changelogdir'
CL_MAX_ENTRIES = 'nsslapd-changelogmaxentries'
CL_MAXAGE = 'nsslapd-changelogmaxage'
CL_COMPACT_INTERVAL = 'nsslapd-changelogcompactdb-interval'
CL_TRIM_INTERVAL = 'nsslapd-changelogtrim-interval'
CL_CONCURRENT_WRITES = 'nsslapd-changelogmaxconcurrentwrites'
CL_ENCRYPT_ALG = 'nsslapd-encryptionalgorithm'
CL_SYM_KEY = 'nsSymmetricKey'

CHANGELOG_PROPNAME_TO_ATTRNAME = {CHANGELOG_NAME: 'cn',
                                  CHANGELOG_DIR: CL_DIR,
                                  CHANGELOG_MAXAGE: CL_MAXAGE,
                                  CHANGELOG_MAXENTRIES: CL_MAX_ENTRIES,
                                  CHANGELOG_TRIM_INTERVAL: CL_TRIM_INTERVAL,
                                  CHANGELOG_COMPACT_INTV: CL_COMPACT_INTERVAL,
                                  CHANGELOG_CONCURRENT_WRITES:
                                      CL_CONCURRENT_WRITES,
                                  CHANGELOG_ENCRYPT_ALG: CL_ENCRYPT_ALG,
                                  CHANGELOG_SYM_KEY: CL_SYM_KEY}

####################################
#
# Properties of an entry
#
####################################
ENTRY_OBJECTCLASS = 'objectclass'
ENTRY_SN = 'sn'
ENTRY_CN = 'cn'
ENTRY_TYPE_PERSON = 'person'
ENTRY_TYPE_INETPERSON = 'inetperson'
ENTRY_TYPE_GROUP = 'group'
ENTRY_USERPASSWORD = 'userpassword'
ENTRY_UID = 'uid'

ENTRY_TYPE_TO_OBJECTCLASS = {ENTRY_TYPE_PERSON: ["top",
                                                 "person"],
                             ENTRY_TYPE_INETPERSON: ["top",
                                                     "person",
                                                     "inetOrgPerson"]}

####################################
#
# Properties supported by the replica agreement
#
####################################

#
# Repl Agreement Attributes
#
AGMT_BIND_METHOD = 'nsDS5ReplicaBindMethod'
AGMT_HOST = 'nsDS5ReplicaHost'
AGMT_PORT = 'nsDS5ReplicaPort'
AGMT_TRANSPORT_INFO = 'nsDS5ReplicaTransportInfo'
AGMT_ATTR_LIST = 'nsDS5ReplicatedAttributeList'
AGMT_ATTR_LIST_TOTAL = 'nsDS5ReplicatedAttributeListTotal'
AGMT_TIMEOUT = 'nsds5replicaTimeout'
AGMT_CRED = 'nsDS5ReplicaCredentials'
AGMT_REFRESH = 'nsds5BeginReplicaRefresh'
AGMT_SCHEDULE = 'nsds5ReplicaUpdateSchedule'
AGMT_BUSY_WAIT_TIME = 'nsds5ReplicaBusyWaitTime'
AGMT_SESSION_PAUSE_TIME = 'nsds5ReplicaSessionPauseTime'
AGMT_ENABLED = 'nsds5ReplicaEnabled'
AGMT_STRIP_ATTRS = 'nsds5ReplicaStripAttrs'
AGMT_FLOW_WINDOW = 'nsds5ReplicaFlowControlWindow'
AGMT_FLOW_PAUSE = 'nsds5ReplicaFlowControlPause'
AGMT_MAXCSN = 'nsds5AgmtMaxCSN'
AGMT_UPDATE_START = 'nsds5replicaLastUpdateStart'
AGMT_UPDATE_END = 'nsds5replicaLastUpdateEnd'
AGMT_CHANGES_SINCE_STARTUP = 'nsds5replicaChangesSentSinceStartup'  # base64
AGMT_UPDATE_STATUS = 'nsds5replicaLastUpdateStatus'
AGMT_UPDATE_STATUS_JSON = 'nsds5replicaLastUpdateStatusJSON'
AGMT_UPDATE_IN_PROGRESS = 'nsds5replicaUpdateInProgress'
AGMT_INIT_START = 'nsds5replicaLastInitStart'
AGMT_INIT_END = 'nsds5replicaLastInitEnd'
AGMT_INIT_STATUS = 'nsds5replicaLastInitStatus'

#
# WinSync Agreement Attributes
#
AGMT_WINSYNC_SUBTREE = 'nsds7WindowsReplicaSubtree'
AGMT_WINSYNC_REPLICA_SUBTREE = 'nsds7DirectoryReplicaSubtree'
AGMT_WINSYNC_NEW_USERSYNC = 'nsds7NewWinUserSyncEnabled'
AGMT_WINSYNC_NEW_GROUP_SYNC = 'nsds7NewWinGroupSyncEnabled'
AGMT_WINSYNC_DOMAIN = 'nsds7WindowsDomain'
AGMT_WINSYNC_COOKIE = 'nsds7DirsyncCookie'
AGMT_WINSYNC_INTERVAL = 'winSyncInterval'
AGMT_WINSYNC_ONE_WAY_SYNC = 'oneWaySync'
AGMT_WINSYNC_MOVE_ACTION = 'winSyncMoveAction'
AGMT_WINSYNC_WIN_FILTER = 'winSyncWindowsFilter'
AGMT_WINSYNC_DIR_FILTER = 'winSyncDirectoryFilter'
AGMT_WINSYNC_SUBTREE_PAIR = 'winSyncSubtreePair'

RA_NAME = 'ReplicaName'
RA_SUFFIX = 'ReplicaRoot'
RA_BINDDN = 'ReplicaBindDN'
RA_BINDPW = 'ReplicaBindCredentials'
RA_METHOD = 'ReplicaBindMethod'
RA_DESCRIPTION = 'description'
RA_SCHEDULE = 'ReplicaSchedule'
RA_TRANSPORT_PROT = 'ReplicaTransportInfo'
RA_FRAC_EXCLUDE = 'ReplicatedAttributeList'
RA_FRAC_EXCLUDE_TOTAL_UPDATE = 'ReplicatedAttributeListTotal'
RA_FRAC_STRIP = 'ReplicaStripAttrs'
RA_CONSUMER_PORT = 'ReplicaPort'
RA_CONSUMER_HOST = 'ReplicaHost'
RA_CONSUMER_TOTAL_INIT = 'ReplicaBeginRefresh'
RA_TIMEOUT = 'ReplicaConnTimeout'
RA_CHANGES = 'ReplicaChangesSentSinceStartup'
RA_BUSY_WAIT = 'ReplicaBusyWaitTime'
RA_PAUSE_TIME = 'ReplicaSessionPauseTime'
RA_ENABLED = 'ReplicaEnabled'
RA_FLOW_WINDOW = 'ReplicaFlowControlWindow'
RA_FLOW_PAUSE = 'ReplicaFlowControlPause'
RA_RUV = 'ReplicaRUV'
RA_MAXCSN = 'ReplicaMaxCSN'
RA_LAST_UPDATE_START = 'ReplicaLastUpdateStart'
RA_LAST_UPDATE_END = 'ReplicaLastUpdateEnd'
RA_LAST_UPDATE_STATUS = 'ReplicaLastUpdateStatus'
RA_UPDATE_IN_PROGRESS = 'ReplicaUpdateInProgress'
RA_LAST_INIT_START = 'ReplicaLastInitStart'
RA_LAST_INIT_END = 'ReplicaLastInitEnd'
RA_LAST_INIT_STATUS = 'ReplicaLastInitStatus'

REPLICA_OBJECTCLASS_VALUE = 'nsds5Replica'
RA_OBJECTCLASS_VALUE = "nsds5replicationagreement"
RA_WINDOWS_OBJECTCLASS_VALUE = "nsDSWindowsReplicationAgreement"

RA_PROPNAME_TO_ATTRNAME = {RA_NAME: 'cn',
                           RA_SUFFIX: REPL_ROOT,
                           RA_BINDDN: REPL_BINDDN,
                           RA_BINDPW: AGMT_CRED,
                           RA_METHOD: AGMT_BIND_METHOD,
                           RA_DESCRIPTION: 'description',
                           RA_SCHEDULE: AGMT_SCHEDULE,
                           RA_TRANSPORT_PROT: AGMT_TRANSPORT_INFO,
                           RA_FRAC_EXCLUDE: AGMT_ATTR_LIST,
                           RA_FRAC_EXCLUDE_TOTAL_UPDATE: AGMT_ATTR_LIST_TOTAL,
                           RA_FRAC_STRIP: AGMT_STRIP_ATTRS,
                           RA_CONSUMER_PORT: AGMT_PORT,
                           RA_CONSUMER_HOST: AGMT_HOST,
                           RA_CONSUMER_TOTAL_INIT: AGMT_REFRESH,
                           RA_TIMEOUT: AGMT_TIMEOUT,
                           RA_CHANGES: AGMT_CHANGES_SINCE_STARTUP,
                           RA_BUSY_WAIT: AGMT_BUSY_WAIT_TIME,
                           RA_PAUSE_TIME: AGMT_SESSION_PAUSE_TIME,
                           RA_ENABLED: AGMT_ENABLED,
                           RA_FLOW_WINDOW: AGMT_FLOW_WINDOW,
                           RA_FLOW_PAUSE: AGMT_FLOW_PAUSE,
                           RA_RUV: REPL_RUV,
                           RA_MAXCSN: AGMT_MAXCSN,
                           RA_LAST_UPDATE_START: AGMT_UPDATE_START,
                           RA_LAST_UPDATE_END: AGMT_UPDATE_END,
                           RA_LAST_UPDATE_STATUS: AGMT_UPDATE_STATUS,
                           RA_UPDATE_IN_PROGRESS: AGMT_UPDATE_IN_PROGRESS,
                           RA_LAST_INIT_START: AGMT_INIT_START,
                           RA_LAST_INIT_END: AGMT_INIT_END,
                           RA_LAST_INIT_STATUS: AGMT_INIT_STATUS,
                           REPLICA_REAP_ACTIVE: REPL_REAP_ACTIVE}

####################################
#
# Properties supported by the plugins
#
####################################

PLUGIN_NAME = "name"
PLUGIN_PATH = "path"
PLUGIN_ENABLE = 'enable'

PLUGINS_OBJECTCLASS_VALUE = "nsSlapdPlugin"
PLUGINS_ENABLE_ON_VALUE = "on"
PLUGINS_ENABLE_OFF_VALUE = "off"

PLUGIN_PROPNAME_TO_ATTRNAME = {PLUGIN_NAME: 'cn',
                               PLUGIN_PATH: 'nsslapd-pluginPath',
                               PLUGIN_ENABLE: 'nsslapd-pluginEnabled'}

####################################
#
# Properties supported by the index
#
####################################
INDEX_TYPE = "type"
INDEX_SYSTEM = "system"
INDEX_MATCHING_RULE = "matching-rule"

INDEX_PROPNAME_TO_ATTRNAME = {INDEX_TYPE: 'nsIndexType',
                              INDEX_SYSTEM: 'nsSystemIndex',
                              INDEX_MATCHING_RULE: 'nsMatchingRule'}

####################################
#
# Properties supported by the tasks
#
####################################

TASK_WAIT = "wait"
TASK_WATCH = "watch"
TASK_TOMB_STRIP = "strip-csn"
EXPORT_REPL_INFO = "repl-info"


####################################
#
# Properties related to logging.
#
####################################

LOG_ACCESS_ENABLED = "nsslapd-accesslog-logging-enabled"
LOG_ACCESS_LEVEL = "nsslapd-accesslog-level"
LOG_ACCESS_PATH = "nsslapd-accesslog"

LOG_ERROR_ENABLED = "nsslapd-errorlog-logging-enabled"
LOG_ERROR_LEVEL = "nsslapd-errorlog-level"
LOG_ERROR_PATH = "nsslapd-errorlog"


def rawProperty(prop):
    '''
        Return the string 'prop' without the heading '+'/'-'
        @param prop - string of a property name

        @return property name without heading '+'/'-'

        @raise None
    '''
    if str(prop).startswith('+') or str(prop).startswith('-'):
        return prop[1::]
    else:
        return prop


def inProperties(prop, properties):
    '''
        Return True if 'prop' is in the 'properties' dictionary
        Properties in 'properties' does NOT contain a heading '+'/'-',
         but 'prop' may contain heading '+'/'-'
        @param prop - string of a property name
        @param properties - dictionary of properties.

        @return True/False

        @raise None
    '''
    if rawProperty(prop) in properties:
        return True
    else:
        return False



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
import json
from lib389.backend import Backends
from lib389.plugins import ManagedEntriesPlugin, MEPConfig, MEPConfigs, MEPTemplate, MEPTemplates
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add
from lib389.cli_base import CustomHelpFormatter

arg_to_attr = {
    'config_area': 'nsslapd-pluginconfigarea'
}

arg_to_attr_config = {
    'scope': 'originScope',
    'filter': 'originFilter',
    'managed_base': 'managedBase',
    'managed_template': 'managedTemplate'
}

arg_to_attr_template = {
    'rdn_attr': 'meprdnattr',
    'static_attr': 'mepstaticattr',
    'mapped_attr': 'mepmappedattr'
}


def mep_edit(inst, basedn, log, args):
    log = log.getChild('mep_edit')
    plugin = ManagedEntriesPlugin(inst)
    generic_object_edit(plugin, log, args, arg_to_attr)


def mep_config_list(inst, basedn, log, args):
    log = log.getChild('mep_config_list')
    plugin = ManagedEntriesPlugin(inst)
    config_area = plugin.get_attr_val_utf8_l('nsslapd-pluginConfigArea')
    configs = MEPConfigs(inst, config_area)
    result = []
    result_json = []
    for config in configs.list():
        if args.json:
            result_json.append(json.loads(config.get_all_attrs_json()))
        else:
            result.append(config.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No Managed Entry plugin instances")


def mep_config_add(inst, basedn, log, args):
    log = log.getChild('mep_config_add')
    plugin = ManagedEntriesPlugin(inst)
    config_area = plugin.get_attr_val_utf8_l('nsslapd-pluginConfigArea')
    if config_area is None:
        config_area = plugin.dn
    props = {'cn': args.NAME}
    generic_object_add(MEPConfig, inst, log, args, arg_to_attr_config, basedn=config_area, props=props)


def mep_config_edit(inst, basedn, log, args):
    log = log.getChild('mep_config_edit')
    plugin = ManagedEntriesPlugin(inst)
    config_area = plugin.get_attr_val_utf8_l('nsslapd-pluginConfigArea')
    configs = MEPConfigs(inst, config_area)
    config = configs.get(args.NAME)
    generic_object_edit(config, log, args, arg_to_attr_config)


def mep_config_show(inst, basedn, log, args):
    log = log.getChild('mep_config_show')
    plugin = ManagedEntriesPlugin(inst)
    config_area = plugin.get_attr_val_utf8_l('nsslapd-pluginConfigArea')
    configs = MEPConfigs(inst, config_area)
    config = configs.get(args.NAME)

    if not config.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.name)
    if args and args.json:
        o_str = config.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(config.display())


def mep_config_del(inst, basedn, log, args):
    log = log.getChild('mep_config_del')
    plugin = ManagedEntriesPlugin(inst)
    config_area = plugin.get_attr_val_utf8_l('nsslapd-pluginConfigArea')
    configs = MEPConfigs(inst, config_area)
    config = configs.get(args.NAME)
    config.delete()
    log.info("Successfully deleted the %s", config.dn)


def mep_template_list(inst, basedn, log, args):
    log = log.getChild('mep_template_list')
    if args.BASEDN is None:
        # Gather all the templates from all the suffixes
        templates = []
        backends = Backends(inst).list()
        for be in backends:
            temps = MEPTemplates(inst, be.get_suffix()).list()
            if len(temps) > 0:
                templates += MEPTemplates(inst, be.get_suffix()).list()
    else:
        templates = MEPTemplates(inst, args.BASEDN).list()

    result = []
    result_json = []
    for template in templates:
        if args.json:
            result_json.append(json.loads(template.get_all_attrs_json()))
        else:
            result.append(template.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No Managed Entry template entries found")


def mep_template_add(inst, basedn, log, args):
    log = log.getChild('mep_template_add')
    targetdn = args.DN
    if not targetdn or not ldap.dn.is_dn(targetdn):
        raise ValueError("Specified DN is not a valid DN")
    generic_object_add(MEPTemplate, inst, log, args, arg_to_attr_template, dn=targetdn)
    log.info('Don\'t forget to assign the template to Managed Entry Plugin config '
             'attribute - managedTemplate')


def mep_template_edit(inst, basedn, log, args):
    log = log.getChild('mep_template_edit')
    targetdn = args.DN
    if not ldap.dn.is_dn(targetdn):
        raise ValueError("Specified DN is not a valid DN")
    template = MEPTemplate(inst, targetdn)
    generic_object_edit(template, log, args, arg_to_attr_template)


def mep_template_show(inst, basedn, log, args):
    log = log.getChild('mep_template_show')
    targetdn = args.DN
    if not ldap.dn.is_dn(targetdn):
        raise ValueError("Specified DN is not a valid DN")
    template = MEPTemplate(inst, targetdn)

    if not template.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % targetdn)
    if args and args.json:
        o_str = template.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(template.display())


def mep_template_del(inst, basedn, log, args):
    log = log.getChild('mep_template_del')
    targetdn = args.DN
    if not ldap.dn.is_dn(targetdn):
        raise ValueError("Specified DN is not a valid DN")
    template = MEPTemplate(inst, targetdn)
    template.delete()
    log.info("Successfully deleted the %s", targetdn)


def _add_parser_args_config(parser):
    parser.add_argument('--scope', help='Sets the scope of the search to use to see '
                                        'which entries the plug-in monitors (originScope)')
    parser.add_argument('--filter', help='Sets the search filter to use to search for and identify the entries '
                                         'within the subtree which require a managed entry (originFilter)')
    parser.add_argument('--managed-base', help='Sets the subtree under which to create '
                                               'the managed entries (managedBase)')
    parser.add_argument('--managed-template', help='Identifies the template entry to use to create '
                                                   'the managed entry (managedTemplate)')


def _add_parser_args_template(parser):
    parser.add_argument('--rdn-attr', help='Sets which attribute to use as the naming attribute '
                                           'in the automatically-generated entry (mepRDNAttr)')
    parser.add_argument('--static-attr', nargs='+',
                        help='Sets an attribute with a defined value that must be added '
                             'to the automatically-generated entry (mepStaticAttr)')
    parser.add_argument('--mapped-attr', nargs='+',
                        help='Sets attributes in the Managed Entries template entry which must exist '
                             'in the generated entry (mepMappedAttr)')


def create_parser(subparsers):
    mep = subparsers.add_parser('managed-entries', help='Manage and configure Managed Entries Plugin', formatter_class=CustomHelpFormatter)
    subcommands = mep.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, ManagedEntriesPlugin)

    edit = subcommands.add_parser('set', help='Edit the plugin settings', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=mep_edit)
    edit.add_argument('--config-area', help='Sets the value of the nsslapd-pluginConfigArea attribute')

    list = subcommands.add_parser('list', help='List Managed Entries Plugin configs and templates', formatter_class=CustomHelpFormatter)
    subcommands_list = list.add_subparsers(help='action')
    list_configs = subcommands_list.add_parser('configs', help='List Managed Entries Plugin configs (list config-area '
                                                               'if specified in the main plugin entry)')
    list_configs.set_defaults(func=mep_config_list)
    list_templates = subcommands_list.add_parser('templates',
                                               help='List Managed Entries Plugin templates in the directory')
    list_templates.add_argument('BASEDN', nargs='?', help='The base DN where to search the templates')
    list_templates.set_defaults(func=mep_template_list)

    config = subcommands.add_parser('config', help='Handle Managed Entries Plugin configs', formatter_class=CustomHelpFormatter)
    config.add_argument('NAME', help='The config entry CN')
    config_subcommands = config.add_subparsers(help='action')
    add = config_subcommands.add_parser('add', help='Add the config entry', formatter_class=CustomHelpFormatter)
    add.set_defaults(func=mep_config_add)
    _add_parser_args_config(add)
    edit = config_subcommands.add_parser('set', help='Edit the config entry', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=mep_config_edit)
    _add_parser_args_config(edit)
    show = config_subcommands.add_parser('show', help='Display the config entry', formatter_class=CustomHelpFormatter)
    show.set_defaults(func=mep_config_show)
    delete = config_subcommands.add_parser('delete', help='Delete the config entry', formatter_class=CustomHelpFormatter)
    delete.set_defaults(func=mep_config_del)

    template = subcommands.add_parser('template', help='Handle Managed Entries Plugin templates', formatter_class=CustomHelpFormatter)
    template.add_argument('DN', help='The template entry DN.')
    template_subcommands = template.add_subparsers(help='action')
    add = template_subcommands.add_parser('add', help='Add the template entry', formatter_class=CustomHelpFormatter)
    add.set_defaults(func=mep_template_add)
    _add_parser_args_template(add)
    edit = template_subcommands.add_parser('set', help='Edit the template entry', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=mep_template_edit)
    _add_parser_args_template(edit)
    show = template_subcommands.add_parser('show', help='Display the template entry', formatter_class=CustomHelpFormatter)
    show.set_defaults(func=mep_template_show)
    delete = template_subcommands.add_parser('delete', help='Delete the template entry', formatter_class=CustomHelpFormatter)
    delete.set_defaults(func=mep_template_del)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

"""Helpers for managing NSS databases in Directory Server
"""

import copy
import os
import re
import socket
import time
import shutil
import logging
import subprocess
import uuid
from typing import Optional
from datetime import datetime, timedelta
from subprocess import check_output, run, PIPE
from lib389.passwd import password_generate
from lib389._mapped_object_lint import DSLint
from lib389.lint import DSCERTLE0001, DSCERTLE0002
from lib389.utils import ensure_str, format_cmd_list, DSVersion, cert_is_ca

KEYBITS = 4096
CA_NAME = 'Self-Signed-CA'
CERT_NAME = 'Server-Cert'
USER_PREFIX = 'user-'
PIN_TXT = 'pin.txt'
PWD_TXT = 'pwdfile.txt'
CERT_SUFFIX = 'O=testing,L=389ds,ST=Queensland,C=AU'
ISSUER = 'CN=ssca.389ds.example.com,%s' % CERT_SUFFIX
SELF_ISSUER = 'CN={HOSTNAME},givenName={GIVENNAME},%s' % CERT_SUFFIX
USER_ISSUER = 'CN={HOSTNAME},%s' % CERT_SUFFIX
VALID = 24 # Months
VALID_MIN = 61  # Days

# My logger
log = logging.getLogger(__name__)


class NssSsl(DSLint):
    def __init__(self, dirsrv=None, dbpassword=None, dbpath=None, uid=None, gid=None):
        self.dirsrv = dirsrv
        self._certdb = dbpath
        if self._certdb is None:
            self._certdb = self.dirsrv.get_cert_dir()
        self.log = log
        if self.dirsrv is not None:
            self.log = self.dirsrv.log
        if dbpassword is None:
            self.dbpassword = password_generate()
        else:
            self.dbpassword = dbpassword

        self.db_files = {group: [f"{self._certdb}/{f}" for f in files]
                         for group, files in {"dbm_backend": ("key3.db", "cert8.db", "secmod.db"),
                                              "sql_backend": ("key4.db", "cert9.db", "pkcs11.txt"),
                                              "support": ("noise.txt", PIN_TXT, PWD_TXT)}.items()}

    @classmethod
    def lint_uid(cls):
        return 'tls'

    def _assert_not_chain(self, pemfile):
        # To work this out, we open the file and count how many
        # begin key and begin cert lines there are. Any more than 1 is bad.
        count = 0
        with open(pemfile, 'r') as f:
            for line in f:
                if line.startswith('-----BEGIN PRIVATE KEY-----') or line.startswith('-----BEGIN CERTIFICATE-----'):
                    count = count + 1
        if count > 1:
            raise ValueError(f"The file {pemfile} may be a chain file. This is not supported. Break out each certificate and key into unique files, and import them individually.")

    def _lint_certificate_expiration(self):
        """Check all the certificates in the db if they will expire within 30 days
        or have already expired.
        """
        cert_list = []
        all_certs = self._rsa_cert_list()
        for cert in all_certs:
            cert_list.append(self.get_cert_details(cert[0]))
        for cert in cert_list:
            cert_date = cert['expires'].split()[0]
            diff_date = datetime.strptime(cert_date, '%Y-%m-%d').date() - datetime.today().date()
            if diff_date < timedelta(days=0):
                # Expired
                report = copy.deepcopy(DSCERTLE0002)
                report['detail'] = report['detail'].replace('CERT', cert['cn'])
                report['check'] = f'tls:certificate_expiration'
                yield report
            elif diff_date < timedelta(days=30):
                # Expiring within 30 days
                report = copy.deepcopy(DSCERTLE0001)
                report['detail'] = report['detail'].replace('CERT', cert['cn'])
                report['check'] = f'tls:certificate_expiration'
                yield report

    def detect_alt_names(self, alt_names=[]):
        """Attempt to determine appropriate subject alternate names for a host.
        Returns the list of names we derive.

        :param alt_names: A list of alternate names.
        :type alt_names: list[str]
        :returns: list[str]
        """
        if self.dirsrv and self.dirsrv.host not in alt_names:
            alt_names.append(ensure_str(self.dirsrv.host))
        if len(alt_names) == 0:
            alt_names.append(ensure_str(socket.gethostname()))
        return alt_names

    def generate_cert_subject(self, alt_names=[]):
        """Return the cert subject we would generate for this host
        from the lib389 self signed process. This is *not* the subject
        of the actual cert, which could be different.

        :param alt_names: Alternative names you want to configure.
        :type alt_names: [str, ]
        :returns: String of the subject DN.
        """
        if self.dirsrv and len(alt_names) > 0:
            return SELF_ISSUER.format(GIVENNAME=self.dirsrv.get_uuid(), HOSTNAME=alt_names[0])
        elif len(alt_names) > 0:
            return SELF_ISSUER.format(GIVENNAME=uuid.uuid4(), HOSTNAME=alt_names[0])
        else:
            return SELF_ISSUER.format(GIVENNAME=uuid.uuid4(), HOSTNAME='lib389host.localdomain')

    def get_server_cert_subject(self, alt_names=[]):
        """Get the server db subject. For now, this uses generate, but later
        we can make this determined from other factors like x509 parsing.

        :returns: str
        """
        alt_names = self.detect_alt_names(alt_names)
        return self.generate_cert_subject(alt_names)

    def _generate_noise(self, fpath):
        noise = password_generate(256)
        prv_mask = os.umask(0o177)
        try:
            with open(fpath, 'w') as f:
                f.write(noise)
        finally:
            prv_mask = os.umask(prv_mask)

    def reinit(self, uid=None, gid=None):
        """
        Re-init (create) the nss db.
        """
        # 48886: The DB that DS ships with is .... well, broken. Purge it!
        assert self.remove_db()

        try:
            os.makedirs(self._certdb)
        except FileExistsError:
            pass

        if self.dirsrv is None:
            # Write a README to let people know what this is
            readme_file = '%s/%s' % (self._certdb, 'README.txt')
            if not os.path.exists(readme_file):
                with open(readme_file, 'w') as f:
                    f.write("""
SSCA - Simple Self-Signed Certificate Authority

This is part of the 389 Directory Server project's lib389 toolkit. It
creates a simple, standalone certificate authority for testing and
development purposes. It's suitable for evaluation and testing purposes
only.
                    """)

        # In the future we may add the needed option to avoid writing the pin
        # files.
        # Write the pin.txt, and the pwdfile.txt
        prv_mask = os.umask(0o177)
        try:
            pin_file = '%s/%s' % (self._certdb, PIN_TXT)
            if not os.path.exists(pin_file):
                with open(pin_file, 'w') as f:
                    f.write('Internal (Software) Token:%s' % self.dbpassword)

            pwd_text_file = '%s/%s' % (self._certdb, PWD_TXT)
            if not os.path.exists(pwd_text_file):
                with open(pwd_text_file, 'w') as f:
                    f.write('%s' % self.dbpassword)

            if uid is not None and gid is not None:
                os.chown(pin_file, uid, gid)
                os.chown(pwd_text_file, uid, gid)
        finally:
            prv_mask = os.umask(prv_mask)

        # Init the db.
        # 48886; This needs to be sql format ...
        cmd = ['/usr/bin/certutil', '-N', '-d', self._certdb, '-f', '%s/%s' % (self._certdb, PWD_TXT),  '-@', '%s/%s' % (self._certdb, PWD_TXT)]
        self._generate_noise('%s/noise.txt' % self._certdb)
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        self.log.debug("nss output: %s", result)

        if uid is not None and gid is not None:
            for file in self.db_files["sql_backend"]:
                os.chown(file, uid, gid)

        return True


    def _db_exists(self, even_partial=False):
        """Check that a nss db exists at the certpath"""

        fn = any if even_partial else all
        if fn(map(os.path.exists, self.db_files["dbm_backend"])) or \
           fn(map(os.path.exists, self.db_files["sql_backend"])):
            return True
        return False

    def remove_db(self):
        """Remove nss db files at the certpath"""

        files = self.db_files["dbm_backend"] + \
                self.db_files["sql_backend"] + \
                self.db_files["support"]

        for file in files:
            try:
                os.remove(file)
            except FileNotFoundError:
                pass

        assert not self._db_exists()
        return True

    def openssl_rehash(self, certdir):
        """
        Compatibly run c_rehash (on old openssl versions) or openssl rehash (on
        new ones). Prefers openssl rehash, because openssl on versions where
        the rehash command doesn't exist, also doesn't correctly set the return
        code. Instead, we parse the output of `openssl version` and try to
        figure out if we have a new enough version to unconditionally run rehash.
        """

        def only_warning(text):
            for line in text.split('\n'):
                if line and not 'warning' in line.lower():
                    return False
            return True

        try:
            openssl_version = check_output(['/usr/bin/openssl', 'version']).decode('utf-8').strip()
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        rehash_available = DSVersion(openssl_version.split(' ')[1]) >= DSVersion('1.1.0')

        if rehash_available:
            cmd = ['/usr/bin/openssl', 'rehash', certdir]
        else:
            cmd = ['/usr/bin/c_rehash', certdir]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            res = run(cmd, text=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            self.log.debug("nss cmd: %s returned %d STDOUT=%s",
                           format_cmd_list(cmd), res.returncode, res.stdout)
            if res.returncode != 1 or not only_warning(res.stdout):
                res.check_returncode()
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.rstrip())

    def create_rsa_ca(self, months=VALID):
        """
        Create a self signed CA.
        """

        # Wait a second to avoid an NSS bug with serial ids based on time.
        time.sleep(1)
        # Create noise.
        self._generate_noise('%s/noise.txt' % self._certdb)
        # Now run the command. Can we do this with NSS native?
        cmd = [
            '/usr/bin/certutil',
            '-S',
            '-n',
            CA_NAME,
            '-s',
            ISSUER,
            '-x',
            '-g',
            '%s' % KEYBITS,
            '-t',
            'CT,,',
            '-v',
            '%s' % months,
            '-2',
            '--keyUsage',
            'certSigning',
            '-d',
            self._certdb,
            '-z',
            '%s/noise.txt' % self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        cmd_input = b'y\n\n'  # responses to certutil questions
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        result = ensure_str(run(cmd, check=True, stderr=PIPE, stdout=PIPE, input=cmd_input).stdout)
        self.log.debug("nss output: %s", result)
        # Now extract the CAcert to a well know place.
        # This allows us to point the cacert dir here and it "just works"
        cmd = [
            '/usr/bin/certutil',
            '-L',
            '-n',
            CA_NAME,
            '-d',
            self._certdb,
            '-a',
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            certdetails = check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        with open('%s/ca.crt' % self._certdb, 'w') as f:
            f.write(ensure_str(certdetails))
        self.openssl_rehash(self._certdb)
        return True

    def rsa_ca_needs_renew(self):
        """Check is our self signed CA is expired or
        will expire less than a minimum period of time (VALID_MIN)
        """

        cmd = [
            '/usr/bin/certutil',
            '-L',
            '-n',
            CA_NAME,
            '-d',
            self._certdb,
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            certdetails = check_output(cmd, stderr=subprocess.STDOUT, encoding='utf-8')
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        end_date_str = certdetails.split("Not After : ")[1].split("\n")[0]
        date_format = '%a %b %d %H:%M:%S %Y'
        end_date = datetime.strptime(end_date_str, date_format)

        if end_date - datetime.now() < timedelta(days=VALID_MIN):
            return True
        else:
            return False

    def renew_rsa_ca(self, months=VALID):
        """Renew the self signed CA."""

        csr_path = os.path.join(self._certdb, 'CA_renew.csr')
        crt_path = '%s/ca.crt' % self._certdb

        # Create noise.
        self._generate_noise('%s/noise.txt' % self._certdb)

        # Generate a CSR for a new CA cert
        cmd = [
            '/usr/bin/certutil',
            '-R',
            '-s',
            ISSUER,
            '-g',
            '%s' % KEYBITS,
            '-k',
            'NSS Certificate DB:%s' % CA_NAME,
            '-d',
            self._certdb,
            '-z',
            '%s/noise.txt' % self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
            '-a',
            '-o', csr_path,
            ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        # Sign the CSR with our old CA
        cmd = [
            '/usr/bin/certutil',
            '-C',
            '-d',
            self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
            '-a',
            '-i', csr_path,
            '-o', crt_path,
            '-c', CA_NAME,
            '--keyUsage',
            'certSigning',
            '-t',
            'CT,,',
            '-v',
            '%s' % months,
            ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        self.openssl_rehash(self._certdb)

        # Import the new CA to our DB instead of the old CA
        cmd = [
            '/usr/bin/certutil',
            '-A',
            '-n', CA_NAME,
            '-t', "CT,,",
            '-a',
            '-i', crt_path,
            '-d', self._certdb,
            '-f', '%s/%s' % (self._certdb, PWD_TXT),
            ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        return crt_path

    def _rsa_cert_list(self):
        cmd = [
            '/usr/bin/certutil',
            '-L',
            '-d',
            self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        # We can skip the first few lines. They are junk
        # IE ['',
        #     'Certificate Nickname                                         Trust Attributes',
        #     '                                                             SSL,S/MIME,JAR/XPI',
        #     '',
        #     'Self-Signed-CA                                               CTu,u,u',
        #     '']
        lines = result.split('\n')[4:-1]
        # Now make the lines usable
        cert_values = []
        for line in lines:
            if line == '':
                continue
            if line == 'Database needs user init':
                # There are no certs, abort...
                return []
            cert_values.append(re.match(r'^(.*[^\s])[\s]+([^\s]+)$', line.rstrip()).groups())
        return cert_values

    def _openssl_get_csr_subject(self, csr_dir, csr_name):
        cmd = [
            '/usr/bin/openssl',
            'req',
            '-subject',
            '-noout',
            '-in',
            '%s/%s'% (csr_dir, csr_name),
        ]
        self.log.debug("cmd: %s", format_cmd_list(cmd))
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        # Parse the subject string from openssl output
        result = result.replace("subject=", "")
        result = result.replace(" ", "").strip()
        result = result.split(',')
        result = result[slice(None, None, -1)]
        result = ','.join([str(elem) for elem in result])
        return result

    def _openssl_get_csr_sub_alt_names(self, csr_dir, csr_name):
        cmd = [
            '/usr/bin/openssl',
            'req',
            '-noout',
            '-text',
            '-in',
            '%s/%s'% (csr_dir, csr_name),
        ]
        self.log.debug("cmd: %s", format_cmd_list(cmd))
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        subaltnames = []
        lines = result.split('\n')
        for line in lines:
            if 'DNS:' in line:
                names = line.split(',')
                for altname in names:
                    altname = altname.replace("DNS:", "")
                    subaltnames.append(altname.replace(",:", "").strip())

        return subaltnames

    def _csr_show(self, name):
        csr_dir = self.dirsrv.get_cert_dir()
        result = ""
        # Display PEM contents of a CSR file
        if name and csr_dir:
            if os.path.exists(csr_dir + "/" + name + ".csr"):
                cmd = [
                    "/usr/bin/sed",
                    "-n",
                    '/BEGIN NEW/,/END NEW/p',
                    csr_dir + "/" + name + ".csr"
                ]
                self.log.debug("cmd: %s", format_cmd_list(cmd))
                try:
                    result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
                except subprocess.CalledProcessError as e:
                    raise ValueError(e.output.decode('utf-8').rstrip())

        return result

    def _csr_list(self, csr_dir=None):
        csr_list = []
        csr_dir = self.dirsrv.get_cert_dir()
        # Search for .csr file extensions in instance config dir
        cmd = [
            '/usr/bin/find',
            csr_dir,
            '-type',
            'f',
            '-name',
            '*.csr',
            '-printf',
            '%f\\n',
        ]
        self.log.debug("cmd: %s", format_cmd_list(cmd))
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        # Bail out if we cant find any .csr files
        if len(result) == 0:
            return []

        # For each .csr file, get last modified time and subject DN
        for csr_file in result.splitlines():
            csr = []
            # Get last modified time stamp
            cmd = [
                '/usr/bin/date',
                '-r',
                '%s/%s'% (csr_dir, csr_file),
                '+%Y-%m-%d %H:%M:%S',
            ]
            try:
                result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
            except subprocess.CalledProcessError as e:
                raise ValueError(e.output.decode('utf-8').rstrip())

            # Add csr modified timestamp
            csr.append(result.strip())
            # Use openssl to get the csr subject DN
            csr.append(self._openssl_get_csr_subject(csr_dir, csr_file))
            # Use openssl to get the csr subject alt host names
            csr.append(self._openssl_get_csr_sub_alt_names(csr_dir, csr_file))
            # Add csr name, without extension
            csr.append(csr_file.rsplit('.', 1)[0])
            csr_list.append(csr)

        return csr_list

    def _rsa_cert_key_exists(self, cert_tuple):
        name = cert_tuple[0]
        cmd = [
            '/usr/bin/certutil',
            '-K',
            '-d',
            self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        lines = result.split('\n')[1:-1]
        for line in lines:
            m = re.match(r'\<(?P<id>.*)\> (?P<type>\w+)\s+(?P<hash>\w+).*:(?P<name>.+)', line)
            if name == m.group('name'):
                return True
        return False

    def _rsa_cert_is_catrust(self, cert_tuple):
        trust_flags = cert_tuple[1]
        (ssl_flag, mime_flag, jar_flag) = trust_flags.split(',')
        return 'C' in ssl_flag

    def _rsa_cert_is_caclienttrust(self, cert_tuple):
        trust_flags = cert_tuple[1]
        (ssl_flag, mime_flag, jar_flag) = trust_flags.split(',')
        return 'T' in ssl_flag

    def _rsa_cert_is_user(self, cert_tuple):
        """
        Check an RSA cert is user trust

        Sadly we can't check for ext key usage, because NSS makes this really hard.
        """
        trust_flags = cert_tuple[1]
        (ssl_flag, mime_flag, jar_flag) = trust_flags.split(',')
        return 'u' in ssl_flag

    def _rsa_ca_exists(self):
        """
        Detect if a self-signed ca exists
        """
        have_ca = False
        cert_list = self._rsa_cert_list()
        for cert in cert_list:
            if self._rsa_cert_key_exists(cert) and self._rsa_cert_is_catrust(cert):
                have_ca = True
        return have_ca

    def _rsa_key_and_cert_exists(self):
        """
        Check if a valid server key and cert pair exist.
        """
        have_cert = False
        cert_list = self._rsa_cert_list()
        for cert in cert_list:
            # This could do a better check for !ca, and server attrs
            if self._rsa_cert_key_exists(cert) and not self._rsa_cert_is_catrust(cert):
                have_cert = True
        return have_cert

    def _rsa_user_exists(self, name):
        """
        Check if a valid server key and cert pair exist for a user.

        we use the format, user-<name>
        """
        have_user = False
        cert_list = self._rsa_cert_list()
        for cert in cert_list:
            if (cert[0] == '%s%s' % (USER_PREFIX, name)):
                if self._rsa_cert_key_exists(cert) and self._rsa_cert_is_user(cert):
                    have_user = True
        return have_user

    def create_rsa_key_and_cert(self, alt_names=[], months=VALID, name=CERT_NAME):
        """
        Create a key and a cert that is signed by the self signed ca

        This will use the hostname from the DS instance, and takes a list of
        extra names to take.
        """

        alt_names = self.detect_alt_names(alt_names)
        subject = self.generate_cert_subject(alt_names)

        # Wait a second to avoid an NSS bug with serial ids based on time.
        time.sleep(1)
        # Create noise.
        self._generate_noise('%s/noise.txt' % self._certdb)
        cmd = [
            '/usr/bin/certutil',
            '-S',
            '-n',
            name,
            '-s',
            subject,
            # We MUST issue with SANs else ldap wont verify the name.
            '-8', ','.join(alt_names),
            '-c',
            CA_NAME,
            '-g',
            '%s' % KEYBITS,
            '-t',
            ',,',
            '-v',
            '%s' % months,
            '-d',
            self._certdb,
            '-z',
            '%s/noise.txt' % self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        self.log.debug("nss output: %s", result)
        return True

    def create_rsa_key_and_csr(self, alt_names=[], subject=None, name=CERT_NAME):
        """Create a new RSA key and the certificate signing request. This
        request can be submitted to a CA for signing. The returned certificate
        can be added with import_rsa_crt.
        """
        csr_path = os.path.join(self._certdb, '%s.csr' % name)

        if len(alt_names) == 0:
            alt_names = self.detect_alt_names(alt_names)
        if subject is None:
            subject = self.generate_cert_subject(alt_names)

        self.log.debug(f"CSR name -> {name}")
        self.log.debug(f"CSR subject -> {subject}")
        self.log.debug(f"CSR alt_names -> {alt_names}")

        # Wait a second to avoid an NSS bug with serial ids based on time.
        time.sleep(1)
        # Create noise.
        self._generate_noise('%s/noise.txt' % self._certdb)

        cmd = [
            '/usr/bin/certutil',
            '-R',
            # We want a dual purposes client and server cert
            '--keyUsage',
            'digitalSignature,nonRepudiation,keyEncipherment,dataEncipherment',
            '--nsCertType',
            'sslClient,sslServer',
            '--extKeyUsage',
            'clientAuth,serverAuth',
            '-s',
            subject,
            # We MUST issue with SANs else ldap wont verify the name.
            '-8', ','.join(alt_names),
            '-g',
            '%s' % KEYBITS,
            '-d',
            self._certdb,
            '-z',
            '%s/noise.txt' % self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
            '-a',
            '-o', csr_path,
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        return csr_path

    def rsa_ca_sign_csr(self, csr_path, months=VALID):
        """ Given a CSR, sign it with our CA certificate (if present). This
        emits a signed certificate which can be imported with import_rsa_crt.
        """
        crt_path = 'crt'.join(csr_path.rsplit('csr', 1))
        ca_path = '%s/ca.crt' % self._certdb

        cmd = [
            '/usr/bin/certutil',
            '-C',
            '-d',
            self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
            '-v',
            '%s' % months,
            '-a',
            '-i', csr_path,
            '-o', crt_path,
            '-c', CA_NAME,
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        return (ca_path, crt_path)

    def import_rsa_crt(self, ca=None, crt=None, name=CERT_NAME):
        """Given a signed certificate from a ca, import the CA and certificate
        to our database.


        """

        assert ca is not None or crt is not None, "At least one parameter should be specified (ca or crt)"

        if ca is not None:
            if not os.path.exists(ca):
                raise ValueError("The certificate file ({}) does not exist".format(ca))
            self._assert_not_chain(ca)

            shutil.copyfile(ca, '%s/ca.crt' % self._certdb)
            self.openssl_rehash(self._certdb)
            cmd = [
                '/usr/bin/certutil',
                '-A',
                '-n', CA_NAME,
                '-t', "CT,,",
                '-a',
                '-i', '%s/ca.crt' % self._certdb,
                '-d', self._certdb,
                '-f',
                '%s/%s' % (self._certdb, PWD_TXT),
            ]
            self.log.debug("nss cmd: %s", format_cmd_list(cmd))
            try:
                check_output(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                raise ValueError(e.output.decode('utf-8').rstrip())

        if crt is not None:
            if not os.path.exists(crt):
                raise ValueError("The certificate file ({}) does not exist".format(crt))
            self._assert_not_chain(crt)
            cmd = [
                '/usr/bin/certutil',
                '-A',
                '-n', name,
                '-t', ",,",
                '-a',
                '-i', crt,
                '-d', self._certdb,
                '-f',
                '%s/%s' % (self._certdb, PWD_TXT),
            ]
            self.log.debug("nss cmd: %s", format_cmd_list(cmd))
            try:
                check_output(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                raise ValueError(e.output.decode('utf-8').rstrip())
            cmd = [
                '/usr/bin/certutil',
                '-V',
                '-d', self._certdb,
                '-n', name,
                '-u', 'YCV'
            ]
            self.log.debug("nss cmd: %s", format_cmd_list(cmd))
            try:
                check_output(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                raise ValueError(e.output.decode('utf-8').rstrip())

    def create_rsa_user(self, name, months=VALID):
        """
        Create a key and cert for a user to authenticate to the directory.

        Name is the uid of the account, and will become the CN of the cert.
        """
        subject = USER_ISSUER.format(HOSTNAME=name)
        if self._rsa_user_exists(name):
            return subject

        # Wait a second to avoid an NSS bug with serial ids based on time.
        time.sleep(1)
        cmd = [
            '/usr/bin/certutil',
            '-S',
            '-n',
            '%s%s' % (USER_PREFIX, name),
            '-s',
            subject,
            '--keyUsage',
            'digitalSignature,nonRepudiation,keyEncipherment,dataEncipherment',
            '--nsCertType',
            'sslClient',
            '--extKeyUsage',
            'clientAuth',
            '-c',
            CA_NAME,
            '-g',
            '%s' % KEYBITS,
            '-t',
            ',,',
            '-v',
            '%s' % months,
            '-d',
            self._certdb,
            '-z',
            '%s/noise.txt' % self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        self.log.debug("nss output: %s", result)
        # Now extract this into PEM files that we can use.
        # pk12util -o user-william.p12 -d . -k pwdfile.txt -n user-william -W ''
        cmd = [
            'pk12util',
            '-d', self._certdb,
            '-o', '%s/%s%s.p12' % (self._certdb, USER_PREFIX, name),
            '-k', '%s/%s' % (self._certdb, PWD_TXT),
            '-n', '%s%s' % (USER_PREFIX, name),
            '-W', '""'
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        # openssl pkcs12 -in user-william.p12 -passin pass:'' -out file.pem -nocerts -nodes
        # Extract the key
        cmd = [
            'openssl',
            'pkcs12',
            '-in', '%s/%s%s.p12' % (self._certdb, USER_PREFIX, name),
            '-passin', 'pass:""',
            '-out', '%s/%s%s.key' % (self._certdb, USER_PREFIX, name),
            '-nocerts',
            '-nodes'
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        # Extract the cert
        cmd = [
            'openssl',
            'pkcs12',
            '-in', '%s/%s%s.p12' % (self._certdb, USER_PREFIX, name),
            '-passin', 'pass:""',
            '-out', '%s/%s%s.crt' % (self._certdb, USER_PREFIX, name),
            '-nokeys',
            '-clcerts',
            '-nodes'
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        # Convert the cert for userCertificate attr
        cmd = [
            'openssl',
            'x509',
            '-inform', 'PEM',
            '-outform', 'DER',
            '-in', '%s/%s%s.crt' % (self._certdb, USER_PREFIX, name),
            '-out', '%s/%s%s.der' % (self._certdb, USER_PREFIX, name),
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        return subject

    def get_rsa_user(self, name):
        """
        Return a dict of information for ca, key and cert paths for the user id
        """
        ca_path = '%s/ca.crt' % self._certdb
        key_path = '%s/%s%s.key' % (self._certdb, USER_PREFIX, name)
        crt_path = '%s/%s%s.crt' % (self._certdb, USER_PREFIX, name)
        crt_der_path = '%s/%s%s.der' % (self._certdb, USER_PREFIX, name)
        return {'ca': ca_path, 'key': key_path, 'crt': crt_path, 'crt_der_path': crt_der_path}

    def list_keys(self, orphan=None):
        key_list = []
        cmd = [
            '/usr/bin/certutil',
            '-K',
            '-d',
            self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        # Ignore the first line of certutil output
        for line in result.splitlines()[1:]:
            # Normalise the output of certutil
            line = re.sub(r"\<[^>]*\>","", line)
            key = re.split(r'\s{2,}', line)
            if orphan:
                if 'orphan' in line:
                    key_list.append(key)
            else:
                key_list.append(key)

        return key_list

    def del_key(self, keyid):
        cmd = [
            '/usr/bin/certutil',
            '-F',
            '-d',
            self._certdb,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
            '-k',
            keyid,
        ]
        try:
            result = ensure_str(check_output(cmd, stderr=subprocess.STDOUT))
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        return result

    # Certificate helper functions
    def del_cert(self,  nickname):
        """Delete this certificate
        """
        cmd = [
                '/usr/bin/certutil',
                '-D',
                '-d', self._certdb,
                '-n', nickname,
                '-f',
                '%s/%s' % (self._certdb, PWD_TXT),
            ]
        self.log.debug("del_cert cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

    def edit_cert_trust(self, nickname,  trust_flags):
        """Edit trust flags
        """

        # validate trust flags
        flag_sections = trust_flags.split(',')
        if len(flag_sections) != 3:
            raise ValueError("Invalid trust flag format")

        for section in flag_sections:
            if len(section) > 6:
                raise ValueError("Invalid trust flag format, too many flags in a section")

        for c in trust_flags:
            if c not in ['p', 'P', 'c',  'C', 'T', 'u', ',']:
                raise ValueError("Invalid trust flag {}".format(c))

        # Modify certificate flags
        cmd = [
            '/usr/bin/certutil',
            '-M',
            '-d', self._certdb,
            '-n', nickname,
            '-t',  trust_flags,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        self.log.debug("edit_cert_trust cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

    def display_cert_details(self, nickname):
        cmd = [
            '/usr/bin/certutil',
            '-d', self._certdb,
            '-n', nickname,
            '-L',
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]
        self.log.debug("display_cert_details cmd: %s", format_cmd_list(cmd))
        try:
            result = check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

        return ensure_str(result)

    def get_cert_details(self, nickname):
        """Get the trust flags, subject DN, issuer, and expiration date

        :return: dict containing certificate details with keys:
            - "cn"          : Certificate nickname (str)
            - "subject"     : Subject DN (str)
            - "issuer"      : Issuer DN (str)
            - "expires"     : Expiration date/time as a string (YYYY-MM-DD HH:MM:SS)
            - "trust_flags" : Trust flags from NSS (str)
        """
        all_certs = self._rsa_cert_list()
        for cert in all_certs:
            if cert[0] == nickname:
                trust_flags = cert[1]
                certdetails = self.display_cert_details(nickname)
                end_date_str = certdetails.split("Not After : ")[1].split("\n")[0]
                date_format = '%a %b %d %H:%M:%S %Y'
                end_date = datetime.strptime(end_date_str, date_format)

                # Subject DN
                subject = ""
                for line in certdetails.splitlines():
                    line = line.lstrip()
                    if line.startswith("Subject: "):
                        subject = line.split("Subject: ")[1].split("\n")[0]
                    elif subject != "":
                        if not line.startswith("Subject Public Key Info:"):
                            subject += line
                        else:
                            # Done, strip off quotes
                            subject = subject[1:-1]
                            break

                # Issuer
                issuer = ""
                for line in certdetails.splitlines():
                    line = line.lstrip()
                    if line.startswith("Issuer: "):
                        issuer = line.split("Issuer: ")[1].split("\n")[0]
                    elif issuer != "":
                        if not line.startswith("Validity:"):
                            issuer += line
                        else:
                            issuer = issuer[1:-1]
                            break

                return {
                    "cn": nickname,
                    "subject": subject,
                    "issuer": issuer,
                    "expires": str(end_date),
                    "trust_flags": trust_flags,
                }

        return None

    def list_certs(self, ca=False):
        all_certs = self._rsa_cert_list()
        certs = []
        for nickname, trust_flags in all_certs:
            if (ca and "CT" in trust_flags) or (not ca and "CT" not in trust_flags):
                cert_details = self.get_cert_details(nickname)
                certs.append(cert_details)
        return certs

    def list_ca_certs(self):
        ca_certs = []
        for cert in self._rsa_cert_list():
            if self._rsa_cert_is_catrust(cert):
                # cert[0] is the nickname
                cert_details = self.get_cert_details(cert[0])
                ca_certs.append(cert_details)
        return ca_certs

    def list_client_ca_certs(self):
        return [
            cert
            for cert in self._rsa_cert_list()
            if self._rsa_cert_is_caclienttrust(cert)
        ]

    def add_cert(self,
                 nickname: str,
                 cert_file: str,
                 pkcs12_password: Optional[str] = None,
                 ca: bool = False,
                 force: bool = False):
        """Add server or CA cert

        :param nickname: Certificate nickname
        :param cert_file: Path to certificate file (PEM, DER, or PKCS#12)
        :param ca: Whether this is a CA certificate
        :param pkcs12_password: Password for PKCS#12, if any
        :param force: Force the addition of a certificate that cannot be verified
        """
        if not nickname:
            raise ValueError("Certificate nickname must not be empty")

        # Verify input_file exists
        if not os.path.isfile(cert_file):
            raise ValueError("The certificate file ({}) does not exist".format(cert_file))

        pem_file = True
        if not cert_file.lower().endswith(".pem"):
            pem_file = False
        else:
            self._assert_not_chain(cert_file)

        if ca:
            # Verify this is a CA cert
            if not cert_is_ca(cert_file):
                raise ValueError(f"Certificate ({nickname}) is not a CA certificate")
            trust_flags = "CT,,"
        else:
            # Verify this is a server cert
            if cert_is_ca(cert_file, pkcs12_password=pkcs12_password):
                raise ValueError(f"Certificate ({nickname}) is not a server certificate")
            trust_flags = ",,"

        pkcs12_file = None
        if cert_file.lower().endswith((".p12", ".pfx")):
            pkcs12_file = True

        if pkcs12_file:
            self.log.info("Importing PKCS#12 into NSS: %s", cert_file)

            if pkcs12_password is None:
                pkcs12_password = ""

            cmd = [
                "pk12util",
                "-v",
                "-n", nickname,
                "-i", cert_file,
                "-d", self._certdb,
                "-k", f"{self._certdb}/{PWD_TXT}",
                "-W", pkcs12_password,
            ]

            # Mask the password in logs
            masked_cmd = [arg if arg != pkcs12_password else "****" for arg in cmd]
            self.log.info(f"nss import p12 cmd: {format_cmd_list(masked_cmd)}", )
            try:
                check_output(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                raise ValueError(f"Failed to import PKCS#12 {cert_file} {e}")

            if ca:
                cmd = [
                    "certutil",
                    "-M",
                    "-n", nickname,
                    "-t", "CT,,",
                    "-d", self._certdb,
                    "-f", f"{self._certdb}/{PWD_TXT}",
                ]
                self.log.debug(f"set CA trust cmd: {format_cmd_list(cmd)}")
                try:
                    check_output(cmd, stderr=subprocess.STDOUT)
                except subprocess.CalledProcessError as e:
                    raise ValueError(f"Failed to set CA trust for {nickname} {e}")

            return

        pem_file = cert_file.lower().endswith(".pem")
        if pem_file:
            self._assert_not_chain(cert_file)

        trust_flags = "CT,," if ca else ",,"

        cmd = [
            '/usr/bin/certutil',
            '-A',
            '-d', self._certdb,
            '-n', nickname,
            '-t', trust_flags,
            '-i', cert_file,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),

        ]

        if pem_file:
            cmd.append('-a')

        self.log.debug("add_cert cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())

    def add_server_key_and_cert(self, input_key, input_cert, name=CERT_NAME):
        if not os.path.exists(input_key):
            raise ValueError("The key file ({}) does not exist".format(input_key))
        if not os.path.exists(input_cert):
            raise ValueError("The cert file ({}) does not exist".format(input_cert))

        self.log.debug(f"Importing key and cert -> {input_key}, {input_cert}")

        p12_bundle = "%s/temp_server_key_cert.p12" % self._certdb

        # Remove the p12 if it exists
        if os.path.exists(p12_bundle):
            os.remove(p12_bundle)

        # Transform to p12
        cmd = [
            'openssl',
            'pkcs12',
            '-export',
            '-in', input_cert,
            '-inkey', input_key,
            '-out', p12_bundle,
            '-name', name,
            '-passout', 'pass:',
            '-aes128'
        ]
        self.log.debug("nss cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
        # Remove the server-cert if it exists, because else the import name fails.
        try:
            self.del_cert(name)
        except:
            pass
        try:
            # Import it
            cmd = [
                'pk12util',
                '-v',
                '-i', p12_bundle,
                '-d', self._certdb,
                '-k', '%s/%s' % (self._certdb, PWD_TXT),
                '-W', "",
            ]
            self.log.debug("nss cmd: %s", format_cmd_list(cmd))
            try:
                check_output(cmd, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as e:
                raise ValueError(e.output.decode('utf-8').rstrip())
        finally:
            # Remove the p12
            if os.path.exists(p12_bundle):
                os.remove(p12_bundle)

    def add_ca_cert(self, cert_file: str, nickname: str, force: bool = False):
        """
        Add a CA certificate from a PEM bundle or single PEM/DER file.

        Adapter function to match abstraction layer interface.

        :param cert_file: path to the certificate file
        :param nickname: nickname to assign
        :param force: Force the addition of a certificate that cannot be verified
        """
        # Verify input_file exists
        if not os.path.exists(cert_file):
            raise ValueError(f"The certificate file ({cert_file}) does not exist")

        # Normalise nickname(s)
        if isinstance(nickname, list):
            nicknames = nickname
        elif isinstance(nickname, str):
            nicknames = [nickname]
        else:
            raise TypeError(f"nickname must be str or list[str], got: {type(nickname)}")

        # Allow overwrite only for single cert
        if len(nicknames) == 1:
            single_nick = nicknames[0]
            try:
                if self.get_cert_details(single_nick):
                    if not force:
                        raise ValueError(
                            f"Certificate already exists with the same name ({single_nick})"
                        )
                    else:
                        log.info(f"Overwriting existing certificate ({single_nick})")
                        self.del_cert(single_nick)
            except ValueError:
                pass

        # Offload to PEM bundle handler
        if cert_file.lower().endswith(".pem"):
            return self.add_ca_cert_bundle(
                cert_file=cert_file,
                nicknames=nicknames
            )

        if len(nicknames) != 1:
            raise ValueError(
                "Binary CA cert requires exactly one nickname"
            )

        if not cert_is_ca(cert_file):
            raise ValueError(f"Certificate ({nickname}) is not a CA certificate")

        self.add_cert(nicknames[0], cert_file, ca=True)

        log.info(f"Successfully added CA certificate ({nickname})")

    def add_ca_cert_bundle(self, cert_file, nicknames):
        """
        Add a PEM file that could be a bundle of CA certs

        :param nicknames: list of names of each CA certificate
        :param cert_file: path to certificate PEM file
        :raises:
        """

        # Verify input_file exists
        if not os.path.exists(cert_file):
            raise ValueError("The certificate file ({}) does not exist".format(cert_file))

        if not cert_file.lower().endswith(".pem"):
            # Binary cert, this can not be a bundle
            self.add_cert(nicknames[0], cert_file, ca=True)
            log.info(f"Successfully added CA certificate ({nicknames[0]})")
            return

        try:
            with open(cert_file, 'r') as f:
                ca_count = 0
                ca_files_to_cleanup = []
                writing_ca = False
                try:
                    for line in f:
                        if line.startswith('-----BEGIN CERTIFICATE-----'):
                            # create tmp cert file
                            writing_ca = True
                            ca_file_name = cert_file + "-" + str(ca_count)
                            ca_file = open(ca_file_name, 'w')
                            ca_files_to_cleanup.append(ca_file_name)
                            ca_file.write(line)
                        elif writing_ca:
                            ca_file.write(line)
                            if line.startswith('-----END CERTIFICATE-----'):
                                writing_ca = False
                                ca_file.close()

                                # Generate CA certificate nickname
                                names_len = len(nicknames)
                                if names_len > ca_count:
                                    if nicknames[ca_count].lower() == CERT_NAME.lower() or nicknames[ca_count].lower() == CA_NAME.lower():
                                        # cleanup
                                        try:
                                            for tmp_file in ca_files_to_cleanup:
                                                os.remove(tmp_file)
                                        except IOError as e:
                                            # failed to remove tmp cert
                                            log.debug("Failed to remove tmp cert file: " + str(e))
                                        raise ValueError(f"You may not import a CA with the nickname {CERT_NAME} or {CA_NAME}")
                                    ca_cert_name = nicknames[ca_count]
                                else:
                                    # A name was not provided for this cert, create one
                                    ca_cert_name = nicknames[names_len - 1] + str(ca_count)

                                # Check if certificate nickname exists
                                cert = self.get_cert_details(ca_cert_name)
                                if cert:
                                    # Not good, cleanup and raise error
                                    try:
                                        for tmp_file in ca_files_to_cleanup:
                                            os.remove(tmp_file)
                                    except IOError as e:
                                        # failed to remove tmp cert
                                        log.debug("Failed to remove tmp cert file: " + str(e))
                                    raise ValueError(f"Certificate already exists with the same name ({ca_cert_name})")

                                # Verify this is a CA cert
                                if not cert_is_ca(ca_file_name):
                                    raise ValueError(f"Certificate ({ca_cert_name}) is not a CA certificate")

                                # Add this CA certificate
                                self.add_cert(ca_cert_name, ca_file_name, ca=True)
                                ca_count += 1
                                log.info(f"Successfully added CA certificate ({ca_cert_name})")

                    # All done, cleanup
                    try:
                        for tmp_file in ca_files_to_cleanup:
                            os.remove(tmp_file)
                    except IOError as e:
                        # failed to remove tmp cert
                        log.debug("Failed to remove tmp cert file: " + str(e))
                except IOError as e:
                    # Some failure, remove all the tmp files
                    ca_file.close()
                    try:
                        for tmp_file in ca_files_to_cleanup:
                            os.remove(tmp_file)
                    except IOError as e:
                        # failed to remove tmp cert
                        log.debug("Failed to remove tmp cert file: " + str(e))
                    raise e
        except EnvironmentError as e:
            raise e

    def export_cert(self, nickname, output_file=None, der_format=False):
        """
        :param nickname: name of certificate
        :param output_file: name for exported certificate
        :param der_format: export certificate in DER/binary format
        :raise ValueError: error
        """
        cmd = [
            '/usr/bin/certutil',
            '-L',
            '-d', self._certdb,
            '-n', nickname,
            '-f',
            '%s/%s' % (self._certdb, PWD_TXT),
        ]

        # Handle args for PEM vs DER options
        if der_format:
            cmd.append('-r')
            if output_file is None:
                output_file = nickname + ".crt"
        else:
            cmd.append('-a')
            if output_file is None:
                output_file = nickname + ".pem"

        # Set output file name
        if not output_file.startswith("/"):
            # Must be full path, otherwise but it in the cert dir
            output_file = f"{self._certdb}/{output_file}"
        cmd.append('-o')
        cmd.append(output_file)

        self.log.debug("export_cert cmd: %s", format_cmd_list(cmd))
        try:
            check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output.decode('utf-8').rstrip())
